@@ -1,6 +1,6 @@
 /* vim:tw=78:ts=8:sw=4:set ft=c:  */
 /*
-    Copyright (C) 2006-2009 Ben Kibbey <bjk@luxsci.net>
+    Copyright (C) 2006-2010 Ben Kibbey <bjk@luxsci.net>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -16,6 +16,10 @@
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02110-1301  USA
 */
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -37,10 +41,6 @@
 #include <sys/select.h>
 #include <libpwmd.h>
 
-#ifdef HAVE_CONFIG_H
-#include <config.h>
-#endif
-
 #ifdef HAVE_ASSUAN_H
 #include <assuan.h>
 #endif
@@ -96,7 +96,7 @@ int pwmd_strerror_r(gpg_error_t code, char *buf, size_t size)
     if (p) {
 	snprintf(buf, size, "%s", p);
 
-	if (strlen(p) > size)
+	if (strlen(p)+1 > size)
 	    return ERANGE;
 
 	return 0;
@@ -117,6 +117,9 @@ gpg_error_t pwmd_init()
 #endif
 #ifdef ENABLE_NLS
     bindtextdomain("libpwmd", LOCALEDIR);
+#endif
+#ifdef WITH_TCP
+    libssh2_init(0);
 #endif
     gpg_err_init();
     assuan_set_malloc_hooks(pwmd_malloc, pwmd_realloc, pwmd_free);
@@ -135,21 +138,14 @@ gpg_error_t _connect_finalize(pwm_t *pwm)
 	return GPG_ERR_EBADFD;
 
     pwm->fd = active[0];
-    fcntl(pwm->fd, F_SETFL, 0);
 #ifdef WITH_PINENTRY
     pwm->pid = -1;
 #endif
     assuan_set_pointer(pwm->ctx, pwm);
 
 #ifdef WITH_TCP
-    // Until X11 forwarding is supported, disable the remote pwmd pinentry.
-    if (pwm->tcp_conn) {
+    if (pwm->tcp_conn)
 	pwm->tcp_conn->state = SSH_NONE;
-	rc = pwmd_command(pwm, NULL, "SET ENABLE_PINENTRY=0");
-
-	if (rc)
-	    return rc;
-    }
 #endif
 
     if (pwm->name) {
@@ -253,6 +249,61 @@ gpg_error_t pwmd_connect_url_async(pwm_t *pwm, const char *url)
     return _pwmd_connect_url(pwm, url, 1);
 }
 
+int read_hook_common(assuan_context_t ctx, assuan_fd_t fd, void *data,
+	size_t len, ssize_t *ret)
+{
+#ifdef WITH_TCP
+    pwm_t *pwm = assuan_get_pointer(ctx);
+
+    if (pwm && pwm->tcp_conn)
+	*ret = read_hook_ssh(pwm, fd, data, len);
+    else {
+#endif
+#ifdef WITH_LIBPTH
+	*ret = pth_recv((int)fd, data, len, 0);
+#else
+	*ret = recv((int)fd, data, len, 0);
+#endif
+#ifdef WITH_TCP
+    }
+#endif
+
+    return *ret >= 0 ? 1 : 0;
+}
+
+int write_hook_common(assuan_context_t ctx, assuan_fd_t fd,
+	const void *data, size_t len, ssize_t *ret)
+{
+#ifdef WITH_TCP
+    pwm_t *pwm = assuan_get_pointer(ctx);
+
+    if (pwm && pwm->tcp_conn)
+	*ret = write_hook_ssh(pwm, fd, data, len);
+    else {
+#endif
+	/* libassuan cannot handle EAGAIN when doing writes. */
+	do {
+#ifdef WITH_LIBPTH
+	    *ret = pth_send((int)fd, data, len, 0);
+#else
+	    *ret = send((int)fd, data, len, 0);
+#endif
+
+	    if (*ret == -1 && errno == EAGAIN) {
+#ifdef WITH_LIBPTH
+		pth_usleep(50000);
+#else
+		usleep(50000);
+#endif
+	    }
+	} while (*ret == -1 && errno == EAGAIN);
+#ifdef WITH_TCP
+    }
+#endif
+
+    return *ret >= 0 ? 1 : 0;
+}
+
 gpg_error_t pwmd_connect(pwm_t *pwm, const char *path)
 {
     char *socketpath = NULL;
@@ -260,6 +311,7 @@ gpg_error_t pwmd_connect(pwm_t *pwm, const char *path)
     struct passwd pw;
     char *pwbuf;
     gpg_error_t rc;
+    struct assuan_io_hooks io_hooks = {read_hook_common, write_hook_common};
 
     if (!pwm)
 	return GPG_ERR_INV_ARG;
@@ -267,7 +319,7 @@ gpg_error_t pwmd_connect(pwm_t *pwm, const char *path)
     pwbuf = _getpwuid(&pw);
 
     if (!pwbuf)
-	return gpg_error_from_errno(errno);
+	return gpg_error_from_syserror();
 
     if (!path || !*path)
 	socketpath = pwmd_strdup_printf("%s/.pwmd/socket", pw.pw_dir);
@@ -277,8 +329,9 @@ gpg_error_t pwmd_connect(pwm_t *pwm, const char *path)
     pwmd_free(pwbuf);
 
     if (!socketpath)
-	return gpg_error_from_errno(ENOMEM);
+	return GPG_ERR_ENOMEM;
 
+    assuan_set_io_hooks(&io_hooks);
     rc = assuan_socket_connect_ext(&ctx, socketpath, -1, 0);
     pwmd_free(socketpath);
 
@@ -295,6 +348,9 @@ static void disconnect(pwm_t *pwm)
 	return;
 
     assuan_disconnect(pwm->ctx);
+#ifdef WITH_TCP
+    _ssh_disconnect(pwm);
+#endif
     pwm->ctx = NULL;
     pwm->fd = -1;
 }
@@ -339,6 +395,9 @@ void pwmd_close(pwm_t *pwm)
     if (pwm->name)
 	pwmd_free(pwm->name);
 
+    if (pwm->cipher)
+	pwmd_free(pwm->cipher);
+
 #ifdef WITH_TCP
     if (pwm->tcp_conn)
 	_free_ssh_conn(pwm->tcp_conn);
@@ -352,6 +411,28 @@ void pwmd_close(pwm_t *pwm)
     pwmd_free(pwm);
 }
 
+#ifdef WITH_TCP
+static gpg_error_t do_async_command(pwm_t *pwm, char **result)
+{
+    pwmd_async_t s;
+    gpg_error_t rc;
+
+    do {
+	s = pwmd_process(pwm, &rc, result);
+
+	if (s != ASYNC_DONE) {
+#ifdef WITH_LIBPTH
+	    pth_usleep(50000);
+#else
+	    usleep(50000);
+#endif
+	}
+    } while (s != ASYNC_DONE);
+
+    return rc;
+}
+#endif
+
 gpg_error_t pwmd_ssh_connect_async(pwm_t *pwm, const char *host, int port,
 	const char *identity, const char *user, const char *known_hosts)
 {
@@ -369,7 +450,11 @@ gpg_error_t pwmd_ssh_connect(pwm_t *pwm, const char *host, int port,
 #ifndef WITH_TCP
     return GPG_ERR_NOT_IMPLEMENTED;
 #else
-    return _do_pwmd_ssh_connect(pwm, host, port, identity, user, known_hosts, 0);
+    gpg_error_t rc;
+
+    rc = _do_pwmd_ssh_connect_async(pwm, host, port, identity, user,
+	    known_hosts, ASYNC_CMD_CONNECT);
+    return rc ? rc : do_async_command(pwm, NULL);
 #endif
 }
 
@@ -379,21 +464,12 @@ gpg_error_t pwmd_get_hostkey(pwm_t *pwm, const char *host, int port,
 #ifndef WITH_TCP
     return GPG_ERR_NOT_IMPLEMENTED;
 #else
-    char *hostkey;
     gpg_error_t rc;
 
-    rc = _do_pwmd_ssh_connect(pwm, host, port, NULL, NULL, NULL, 1);
+    rc = _do_pwmd_ssh_connect_async(pwm, host, port, NULL, NULL, NULL,
+	    ASYNC_CMD_HOSTKEY);
 
-    if (rc)
-	return rc;
-
-    hostkey = pwmd_strdup(pwm->tcp_conn->hostkey);
-
-    if (!hostkey)
-	rc = gpg_error_from_errno(ENOMEM);
-
-    *result = hostkey;
-    return rc;
+    return rc ? rc : do_async_command(pwm, result);
 #endif
 }
 
@@ -416,7 +492,7 @@ static int inquire_realloc_cb(void *data, const void *buffer, size_t len)
 	return 0;
 
     if ((p = pwmd_realloc(mem->buf, mem->len + len)) == NULL)
-	return gpg_error_from_errno(ENOMEM);
+	return GPG_ERR_ENOMEM;
 
     mem->buf = p;
     memcpy((char *)mem->buf + mem->len, buffer, len);
@@ -428,11 +504,6 @@ static int inquire_cb(void *data, const char *keyword)
 {
     pwm_t *pwm = (pwm_t *)data;
     gpg_error_t rc = 0;
-#ifdef WITH_LIBPTH
-    int flags = pth_fdmode(pwm->fd, PTH_FDMODE_POLL);
-#else
-    int flags = fcntl(pwm->fd, F_GETFL);
-#endif
 
     /* Shouldn't get this far without a callback. */
     if (!pwm->inquire_func)
@@ -446,11 +517,28 @@ static int inquire_cb(void *data, const char *keyword)
 	rc = pwm->inquire_func(pwm->inquire_data, keyword, rc, &result, &len);
 	rc = gpg_err_code(rc);
 
+	if (rc == GPG_ERR_CANCELED) {
+	    rc = assuan_write_flush(pwm->ctx);
+
+	    if (!rc) {
+		rc = assuan_write_line(pwm->ctx, "CANCEL");
+
+		if (!rc)
+		    rc = GPG_ERR_CANCELED;
+	    }
+
+	    break;
+	}
+
 	if (rc == GPG_ERR_EOF || !rc) {
-	    if (len <= 0 || !result) {
+	    if (len <= 0 && !result) {
 		rc = 0;
 		break;
 	    }
+	    else if ((len <= 0 && result) || (len && !result)) {
+		rc = GPG_ERR_INV_ARG;
+		break;
+	    }
 
 	    arc = assuan_send_data(pwm->ctx, result, len);
 	    arc = gpg_err_code(arc);
@@ -466,26 +554,26 @@ static int inquire_cb(void *data, const char *keyword)
 	    break;
 
 	if (!rc) {
-	    /* Set to non-blocking so _pwmd_process() can return. */
-#ifdef WITH_LIBPTH
-	    pth_fdmode(pwm->fd, PTH_FDMODE_NONBLOCK);
-#else
-	    fcntl(pwm->fd, F_SETFL, O_NONBLOCK);
-#endif
+	    pwm->inquire_sent += len;
+
+	    if (pwm->status_func) {
+		char buf[ASSUAN_LINELENGTH];
+
+		snprintf(buf, sizeof(buf), "XFER %u %u", pwm->inquire_sent,
+			pwm->inquire_total);
+		rc = pwm->status_func(pwm->status_data, buf);
+
+		if (rc)
+		    continue;
+	    }
+
 	    rc = _pwmd_process(pwm);
-#ifdef WITH_LIBPTH
-	    pth_fdmode(pwm->fd, flags);
-#else
-	    fcntl(pwm->fd, F_SETFL, flags);
-#endif
+
+	    if (rc == GPG_ERR_EAGAIN)
+		rc = 0;
 	}
     }
 
-#ifdef WITH_LIBPTH
-    pth_fdmode(pwm->fd, flags);
-#else
-    fcntl(pwm->fd, F_SETFL, flags);
-#endif
     return gpg_err_code(rc);
 }
 
@@ -505,7 +593,7 @@ static gpg_error_t do_nb_command(pwm_t *pwm, const char *cmd, ...)
     buf = pwmd_malloc(ASSUAN_LINELENGTH+1);
 
     if (!buf)
-	return gpg_error_from_errno(ENOMEM);
+	return GPG_ERR_ENOMEM;
 
     va_start(ap, cmd);
     len = vsnprintf(buf, ASSUAN_LINELENGTH+1, cmd, ap);
@@ -549,7 +637,7 @@ gpg_error_t pwmd_open_async(pwm_t *pwm, const char *filename)
 	pwm->filename = pwmd_strdup(filename);
 
 	if (!pwm->filename)
-	    return gpg_error_from_errno(ENOMEM);
+	    return GPG_ERR_ENOMEM;
 
 	gpg_error_t rc = send_pinentry_options(pwm);
 
@@ -582,7 +670,18 @@ gpg_error_t pwmd_open_async(pwm_t *pwm, const char *filename)
 	return GPG_ERR_INV_STATE;
 
     pwm->cmd = ASYNC_CMD_OPEN;
-    return do_nb_command(pwm, "OPEN %s %s", f, p ? p : "");
+#ifdef WITH_TCP
+    return do_nb_command(pwm, "OPEN %s%s%s%s %s",
+	    (pwm->opts & OPT_BASE64) ? "--base64 " : "",
+	    (pwm->opts & OPT_LOCK_ON_OPEN) ? "--lock " : "",
+	    pwm->tcp_conn ? "--pinentry=0 " : "",
+	    f, p ? p : "");
+#else
+    return do_nb_command(pwm, "OPEN %s%s%s %s",
+	    (pwm->opts & OPT_BASE64) ? "--base64 " : "",
+	    (pwm->opts & OPT_LOCK_ON_OPEN) ? "--lock " : "",
+	    f, p ? p : "");
+#endif
 }
 
 gpg_error_t pwmd_save_async(pwm_t *pwm)
@@ -615,7 +714,20 @@ gpg_error_t pwmd_save_async(pwm_t *pwm)
 #endif
 
     pwm->cmd = ASYNC_CMD_SAVE;
-    return do_nb_command(pwm, "SAVE %s", p ? p : "");
+
+    char *tmp = pwmd_strdup_printf("%li", pwm->iterations);
+    gpg_error_t rc = do_nb_command(pwm, "SAVE %s%s--iterations=%s --cipher=%s %s",
+	    (pwm->opts & OPT_BASE64) ? "--base64 " : "",
+#ifdef WITH_TCP
+	    pwm->tcp_conn || pwm->lastcmd == ASYNC_CMD_SAVE2 ?  "--pinentry=0 " : "",
+#else
+	    pwm->lastcmd == ASYNC_CMD_SAVE2 ?  "--pinentry=0 " : "",
+#endif
+	    pwm->iterations != -1 ? tmp : "",
+	    pwm->cipher ? pwm->cipher : "", p ? p : "");
+
+    pwmd_free(tmp);
+    return rc;
 }
 
 static gpg_error_t parse_assuan_line(pwm_t *pwm)
@@ -678,6 +790,9 @@ static pwmd_async_t reset_async(pwm_t *pwm, int done)
     }
 #endif
 #ifdef WITH_TCP
+    if (pwm->tcp_conn)
+	pwm->tcp_conn->rc = 0;
+
     if (done && pwm->tcp_conn) {
 	_free_ssh_conn(pwm->tcp_conn);
 	pwm->tcp_conn = NULL;
@@ -871,7 +986,7 @@ pwmd_async_t pwmd_process(pwm_t *pwm, gpg_error_t *rc, char **result)
 		return reset_async(pwm, 1);
 	    }
 	    
-	    pwm->tcp_conn->state = SSH_INIT;
+	    pwm->tcp_conn->state = SSH_NONE;
 	    pwm->tcp_conn->rc = 0;
 	    *rc = _setup_ssh_session(pwm);
 
@@ -893,6 +1008,9 @@ pwmd_async_t pwmd_process(pwm_t *pwm, gpg_error_t *rc, char **result)
 	    case SSH_AUTH:
 		*rc = _setup_ssh_auth(pwm);
 		break;
+	    case SSH_AGENT:
+		*rc = _setup_ssh_agent(pwm);
+		break;
 	    case SSH_CHANNEL:
 		*rc = _setup_ssh_channel(pwm);
 		break;
@@ -911,7 +1029,10 @@ pwmd_async_t pwmd_process(pwm_t *pwm, gpg_error_t *rc, char **result)
 	if (!*rc) {
 	    switch (pwm->tcp_conn->cmd) {
 		case ASYNC_CMD_HOSTKEY:
-		    *result = pwm->result;
+		    *result = pwmd_strdup(pwm->tcp_conn->hostkey);
+
+		    if (!*result)
+			*rc = GPG_ERR_ENOMEM;
 		    break;
 		default:
 		    break;
@@ -973,7 +1094,7 @@ pwmd_async_t pwmd_process(pwm_t *pwm, gpg_error_t *rc, char **result)
 	    pwm->lastcmd = c;
 
 	    if (!pwm->_password) {
-		*rc = gpg_error_from_errno(ENOMEM);
+		*rc = GPG_ERR_ENOMEM;
 		return reset_async(pwm, 0);
 	    }
 
@@ -1095,6 +1216,7 @@ gpg_error_t pwmd_inquire(pwm_t *pwm, const char *cmd, pwmd_inquire_cb_t fn,
 
     pwm->inquire_func = fn;
     pwm->inquire_data = data;
+    pwm->inquire_sent = 0;
     return _assuan_command(pwm, pwm->ctx, NULL, cmd);
 }
 
@@ -1121,7 +1243,7 @@ gpg_error_t pwmd_command_ap(pwm_t *pwm, char **result, const char *cmd,
 
     if (!buf) {
 	va_end(ap2);
-	return gpg_error_from_errno(ENOMEM);
+	return GPG_ERR_ENOMEM;
     }
 
     len = vsnprintf(buf, len, cmd, ap2);
@@ -1302,6 +1424,7 @@ static gpg_error_t do_pwmd_open(pwm_t *pwm, const char *filename, int nb,
     char *result = NULL;
     char *password = NULL;
     gpg_error_t rc;
+    int no_pinentry = 0;
 
     if (pwm->lastcmd != ASYNC_CMD_OPEN2)
 	pwm->pin_try = 0;
@@ -1338,12 +1461,8 @@ static gpg_error_t do_pwmd_open(pwm_t *pwm, const char *filename, int nb,
 #ifdef WITH_PINENTRY
     if (rc == GPG_ERR_NOT_FOUND && local_pinentry) {
 	/* Prevent pwmd from using it's pinentry if the passphrase fails. */
-	if (!pwm->pin_try) {
-	    rc = pwmd_command(pwm, NULL, "SET ENABLE_PINENTRY=0");
-
-	    if (rc)
-		return rc;
-	}
+	if (!pwm->pin_try)
+	    no_pinentry = 1;
 
 	rc = _pinentry_open(pwm, filename, &password, nb);
 
@@ -1368,8 +1487,11 @@ gotpassword:
 	    return rc;
     }
 
-    rc = pwmd_command(pwm, NULL, "OPEN %s %s", filename,
-	    password ? password : "");
+    rc = pwmd_command(pwm, NULL, "OPEN %s%s%s%s %s",
+	    (pwm->opts & OPT_BASE64) ? "--base64 " : "",
+	    (pwm->opts & OPT_LOCK_ON_OPEN) ? "--lock " : "",
+	    no_pinentry ? "--pinentry=0 " : "",
+	    filename, password ?  password : "");
 
     /*
      * Keep the user defined password set with pwmd_setopt(). The password may
@@ -1475,6 +1597,105 @@ gpg_error_t pwmd_open_async2(pwm_t *pwm, const char *filename)
 #endif
 }
 
+struct open_inquire_s {
+    pwm_t *pwm;
+    char *filename;
+    pwmd_inquire_cb_t func;
+    void *user;
+    int sent;
+};
+
+/* We need to send the filename argument here rather than confusing the app.
+ * The app will only send the passphrase and nothing else.
+ */
+static gpg_error_t open_inquire_cb(void *user, const char *cmd,
+	gpg_error_t rc, char **data, size_t *len)
+{
+    struct open_inquire_s *o = user;
+
+    if (rc)
+	return rc;
+
+    if (!o->sent) {
+	char *line = pwmd_strdup_printf("%s ", o->filename);
+	rc = assuan_send_data(o->pwm->ctx, line, strlen(line));
+
+	pwmd_free(line);
+
+	/* Give the user a chance to cleanup. */
+	if (rc)
+	    return o->func(o->user, cmd, rc, data, len);
+
+	o->sent = 1;
+    }
+
+    return o->func(o->user, cmd, rc, data, len);
+}
+
+gpg_error_t pwmd_open_inquire(pwm_t *pwm, const char *filename,
+	pwmd_inquire_cb_t func, void *user)
+{
+    if (!pwm || !filename || !func)
+	return GPG_ERR_INV_ARG;
+
+    if (!pwm->ctx)
+	return GPG_ERR_INV_STATE;
+
+    struct open_inquire_s *inq = pwmd_calloc(1, sizeof(struct open_inquire_s));
+    gpg_error_t rc;
+
+    if (!inq)
+	return GPG_ERR_ENOMEM;
+
+    inq->pwm = pwm;
+    inq->filename = pwmd_strdup(filename);
+    inq->func = func;
+    inq->user = user;
+
+    char *cmd = pwmd_strdup_printf("OPEN --inquire %s%s",
+	    (pwm->opts & OPT_LOCK_ON_OPEN) ? "--lock " : "",
+	    (pwm->opts & OPT_BASE64) ? "--base64 " : "");
+
+    rc = pwmd_inquire(pwm, cmd, open_inquire_cb, inq);
+    pwmd_free(cmd);
+    pwmd_free(inq->filename);
+    pwmd_free(inq);
+
+    if (!rc) {
+	if (pwm->filename)
+	    pwmd_free(pwm->filename);
+
+	pwm->filename = pwmd_strdup(filename);
+
+	if (!pwm->filename) {
+	    pwmd_close(pwm);
+	    rc = GPG_ERR_ENOMEM;
+	}
+    }
+
+    return rc;
+}
+
+gpg_error_t pwmd_save_inquire(pwm_t *pwm, pwmd_inquire_cb_t func, void *user)
+{
+    if (!pwm || !func)
+	return GPG_ERR_INV_ARG;
+
+    if (!pwm->ctx)
+	return GPG_ERR_INV_STATE;
+
+    gpg_error_t rc;
+    char *tmp = pwmd_strdup_printf("%li", pwm->iterations);
+    char *cmd = pwmd_strdup_printf("SAVE %s --inquire --iterations=%s --cipher=%s",
+	    (pwm->opts & OPT_BASE64) ? "--base64 " : "",
+	    pwm->iterations != -1 ? tmp : "",
+	    pwm->cipher ? pwm->cipher : "");
+
+    rc = pwmd_inquire(pwm, cmd, func, user);
+    pwmd_free(cmd);
+    return rc;
+}
+
 static gpg_error_t do_pwmd_save(pwm_t *pwm, int nb, int local_pinentry)
 {
     char *result = NULL;
@@ -1582,7 +1803,24 @@ gotpassword:
 	    return rc;
     }
 
-    rc = pwmd_command(pwm, NULL, "SAVE %s", password ? password : "");
+    char *tmp = pwmd_strdup_printf("%li", pwm->iterations);
+
+#ifdef WITH_TCP
+    rc = pwmd_command(pwm, NULL, "SAVE %s%s--iterations=%s --cipher=%s %s",
+	    (pwm->opts & OPT_BASE64) ? "--base64 " : "",
+	    local_pinentry || pwm->tcp_conn ? "--pinentry=0 " : "",
+	    pwm->iterations != -1 ? tmp : "",
+	    pwm->cipher ? pwm->cipher : "",
+	    password ? password : "");
+#else
+    rc = pwmd_command(pwm, NULL, "SAVE %s--iterations=%s --cipher=%s %s",
+	    (pwm->opts & OPT_BASE64) ? "--base64 " : "",
+	    pwm->iterations != -1 ? tmp : "",
+	    pwm->cipher ? pwm->cipher : "",
+	    password ? password : "");
+#endif
+
+    pwmd_free(tmp);
 
     if (!pwm->passfunc && password && password != pwm->password)
 	pwmd_free(password);
@@ -1632,7 +1870,7 @@ gpg_error_t pwmd_save(pwm_t *pwm)
 gpg_error_t pwmd_setopt(pwm_t *pwm, pwmd_option_t opt, ...)
 {
     va_list ap;
-    int n = va_arg(ap, int);
+    int n;
     char *arg1;
     gpg_error_t rc = 0;
 
@@ -1642,6 +1880,52 @@ gpg_error_t pwmd_setopt(pwm_t *pwm, pwmd_option_t opt, ...)
     va_start(ap, opt);
 
     switch (opt) {
+	case PWMD_OPTION_BASE64:
+	    n = va_arg(ap, int);
+
+	    if (n < 0 || n > 1) {
+		rc = GPG_ERR_INV_VALUE;
+		break;
+	    }
+
+	    if (n)
+		pwm->opts |= OPT_BASE64;
+	    else
+		pwm->opts &= ~OPT_BASE64;
+
+	    break;
+	case PWMD_OPTION_ITERATIONS:
+	    pwm->iterations = va_arg(ap, long);
+
+	    if (pwm->iterations < -1) {
+		pwm->iterations = -1;
+		rc = GPG_ERR_INV_VALUE;
+	    }
+
+	    break;
+	case PWMD_OPTION_CIPHER:
+	    arg1 = va_arg(ap, char *);
+
+	    if (pwm->cipher)
+		pwmd_free(pwm->cipher);
+
+	    pwm->cipher = arg1 ? pwmd_strdup(arg1) : NULL;
+	    break;
+	case PWMD_OPTION_LOCK_ON_OPEN:
+	    n = va_arg(ap, int);
+
+	    if (n < 0 || n > 1)
+		rc = GPG_ERR_INV_VALUE;
+
+	    if (n)
+		pwm->opts |= OPT_LOCK_ON_OPEN;
+	    else
+		pwm->opts &= ~OPT_LOCK_ON_OPEN;
+
+	    break;
+	case PWMD_OPTION_INQUIRE_TOTAL:
+	    pwm->inquire_total = va_arg(ap, size_t);
+	    break;
 	case PWMD_OPTION_STATUS_CB:
 	    pwm->status_func = va_arg(ap, pwmd_status_cb_t);
 	    break;
@@ -1665,22 +1949,20 @@ gpg_error_t pwmd_setopt(pwm_t *pwm, pwmd_option_t opt, ...)
 	case PWMD_OPTION_PINENTRY_TRIES:
 	    n = va_arg(ap, int);
 
-	    if (n <= 0) {
-		va_end(ap);
+	    if (n <= 0)
 		rc = GPG_ERR_INV_VALUE;
-	    }
 	    else
 		pwm->pinentry_tries = n;
+
 	    break;
 	case PWMD_OPTION_PINENTRY_TIMEOUT:
 	    n = va_arg(ap, int);
 
-	    if (n < 0) {
-		va_end(ap);
+	    if (n < 0)
 		rc = GPG_ERR_INV_VALUE;
-	    }
 	    else
 		pwm->pinentry_timeout = n;
+
 	    break;
 	case PWMD_OPTION_PINENTRY_PATH:
 	    if (pwm->pinentry_path)
@@ -1744,8 +2026,8 @@ gpg_error_t pwmd_setopt(pwm_t *pwm, pwmd_option_t opt, ...)
 
 	    pwm->lcmessages = arg1 ? pwmd_strdup(arg1) : NULL;
 	    break;
-	case PWMD_OPTION_IP_VERSION:
 #ifdef WITH_TCP
+	case PWMD_OPTION_IP_VERSION:
 	    n = va_arg(ap, int);
 
 	    switch (n) {
@@ -1758,12 +2040,30 @@ gpg_error_t pwmd_setopt(pwm_t *pwm, pwmd_option_t opt, ...)
 		    rc = GPG_ERR_INV_VALUE;
 		    break;
 	    }
-
-	    va_end(ap);
-#else
-	    rc = GPG_ERR_NOT_IMPLEMENTED;
-#endif
 	    break;
+	case PWMD_OPTION_KNOWNHOST_CB:
+	    pwm->kh_cb = va_arg(ap, pwmd_knownhost_cb_t);
+	    break;
+	case PWMD_OPTION_KNOWNHOST_DATA:
+	    pwm->kh_data = va_arg(ap, void *);
+	    break;
+	case PWMD_OPTION_SSH_AGENT:
+	    pwm->use_agent = va_arg(ap, int);
+
+	    if (pwm->use_agent < 0 || pwm->use_agent > 1) {
+		pwm->use_agent = 0;
+		rc = GPG_ERR_INV_VALUE;
+	    }
+
+	    break;
+#else
+	case PWMD_OPTION_IP_VERSION:
+	case PWMD_OPTION_KNOWNHOST_CB:
+	case PWMD_OPTION_KNOWNHOST_DATA:
+	case PWMD_OPTION_SSH_AGENT:
+	    rc = GPG_ERR_NOT_IMPLEMENTED;
+	    break;
+#endif
 	default:
 	    rc = GPG_ERR_UNKNOWN_OPTION;
 	    break;
@@ -1888,6 +2188,7 @@ pwm_t *pwmd_new(const char *name)
     reset_handle(h);
     h->pinentry_timeout = -30;
     h->pinentry_tries = 3;
+    h->iterations = -1;
 #ifdef WITH_TCP
     h->prot = PWMD_IP_ANY;
 #endif
@@ -1979,3 +2280,8 @@ gpg_error_t pwmd_getpin(pwm_t *pwm, const char *filename, char **result,
     return _pwmd_getpin(pwm, filename, result, which);
 #endif
 }
+
+const char *pwmd_version()
+{
+    return LIBPWMD_VERSION_STR;
+}
