@@ -22,32 +22,46 @@
  * server. Pwmd version 2.0 or later is required; either locally or remotely. 
  */
 
+/*! \section threads Multiple Threads
+ *
+ * \ref libpwmd should be thread-safe on a per handle bases. Meaning that only
+ * one thread should access a \ref pwm_t handle at a time.
+ *
+ * There is a limitation that when compiled with \ref pth(3) support SSH
+ * connections cannot make use of the \ref pth(3) IO functions and therefore
+ * will block while waiting on IO operations in other threads (as of libssh2
+ * 1.2.8).
+ */
+
 /*! \section ssh SSH Details
  *
  * A remote connection to a pwmd server is possible by using an SSH channel
  * which spawns a shell and executes a proxy server that connects to the pwmd
- * local UNIX domain socket. Authentication is done by using SSH public key
+ * local Unix Domain Socket. Authentication is done by using SSH public key
  * (see \ref ssh-keygen(1)) authentication and verifying the host key against
- * a local file containing SHA1 hashes of known hosts. It's a lot like how the
- * standard OpenSSH does things only the known_hosts file is in a different
- * format.
+ * a local OpenSSH known hosts formatted file.
  *
- * The server hash can be had by using \ref pwmd_get_hostkey() and storing the
- * result in a file. This file is then used as the \a known_hosts argument to
- * the SSH connection functions.
+ * The servers public key can be had by using \ref pwmd_get_hostkey() and
+ * storing the result in a file or done automatially by using a callback
+ * function \ref pwmd_knownhost_cb_t while connecting to the unknown host.
  *
- * Here's an example \ref authorized_keys(5) entry. The hash portion should be
- * the same as the contents of the generated \ref ssh-keygen(1) \a
- * identity.pub file which is passed as a parameter to the SSH connection
- * functions:
+ * On the server side you'll need a proxy server to connect to the real pwmd
+ * server. Here's an example \ref authorized_keys(5) entry that will do this.
+ * The \a public_key portion should be the same as the contents of the
+ * generated \ref ssh-keygen(1) \a identity.pub file which is passed as a
+ * parameter to the SSH connection functions:
  *
  * \code
- * command="socat UNIX-CONNECT:$HOME/.pwmd/socket -" <hash> ...
+ * command="socat UNIX-CONNECT:$HOME/.pwmd/socket -" <public_key> ...
  * \endcode
  *
  * \note Only an SSH identity without a passphrase is supported. For now
  * anyway. This is a limitation of libssh2 (version 1.1 as of this writing).
  *
+ * \version 6.0.3
+ * The first version to use the OpenSSH known hosts file format exclusively.
+ * Earlier versions used only an SHA1 hash of the host key.
+ *
  * \x11
  */
 
@@ -111,54 +125,19 @@
  * codes can be described by using \ref pwmd_strerror(), or the thread-safe
  * \ref pwmd_strerror_r().
  *
- * \note Internally, some error codes are a bitmask of an error source. In
- * order to simplify the result codes, libpwmd strips any error source from
- * the error code before returning it.
- */
-
-/*! \section example Example Client
- *
- * The following example will list the element tree of the data file specified
- * in the first command line argument.
- *
- * \code
- * #include <stdio.h>
- * #include <stdlib.h>
- * #include <libpwmd.h>
- *
- * int main(int argc, char **argv)
- * {
- *     pwm_t *pwm = pwmd_new(NULL);
- *     gpg_error_t rc = pwmd_connect(pwm, NULL);
- *     char *result;
- *     
- *     if (!rc) {
- *         rc = pwmd_open(pwm, argv[1]);
- *
- *         if (!rc) {
- *             rc = pwmd_command(pwm, &result, "%s", "LIST");
- *
- *             if (!rc) {
- *                 printf("%s", result);
- *                 pwmd_free(result);
- *             }
- *         }
- *     }
- *
- *     pwmd_close(pwm);
- *     
- *     if (rc)
- *         fprintf(stderr, "ERR: %s\n", pwmd_strerror(rc));
- *
- *     exit(rc ? 1 : 0);
- * }
- * \endcode
+ * \note libgpg-error normally returns an error code as a bitmask of an error
+ * source and the actual error code. In order to simplify the result codes,
+ * libpwmd strips any error source from the error code before returning it.
  */
 
 /*! \file */
 #ifndef LIBPWMD_H
 #define LIBPWMD_H
 
+#ifndef LIBPWMD_API
+#define LIBPWMD_API
+#endif
+
 #include <gpg-error.h>
 #include <stdarg.h>
 
@@ -167,14 +146,56 @@ extern "C" {
 #endif
 
 /*! \def LIBPWMD_VERSION
+ * \brief Version information.
  *
  * The version of this library.
  */
-#define LIBPWMD_VERSION 0x602
+#define LIBPWMD_VERSION 0x603
+
+
+/*! \def LIBPWMD_VERSION_MAJOR
+ * \brief Version information.
+ *
+ * The major release number of this library.
+ */
+#define LIBPWMD_VERSION_MAJOR 6
+
+
+/*! \def LIBPWMD_VERSION_MINOR
+ * \brief Version information.
+ *
+ * The minor release number of this library.
+ */
+#define LIBPWMD_VERSION_MINOR 0
+
+
+/*! \def LIBPWMD_VERSION_PATCH
+ * \brief Version information.
+ *
+ * The patch level of this library.
+ */
+#define LIBPWMD_VERSION_PATCH 3
+
+
+/*! \def LIBPWMD_VERSION_STR
+ * \brief Version information.
+ *
+ * A string representation of the version of this library.
+ */
+#define LIBPWMD_VERSION_STR "6.0.3"
+
+
+/*! \brief Returns this version of libpwmd.
+ *
+ * As a string.
+ * \return A string.
+ */
+LIBPWMD_API const char *pwmd_version();
 
 
 struct pwm_s;
 /*! \typedef pwm_t
+ * \brief libpwmd handle.
  *
  * When a handle is mentioned in this documentation it is a pointer of this
  * type. A new handle is created with \ref pwmd_new().
@@ -183,6 +204,7 @@ typedef struct pwm_s pwm_t;
 
 
 /*! \typedef pwmd_async_t
+ * \brief Asynchronous return value.
  *
  * The return code of \ref pwmd_process() which is used for all asynchronous
  * commands.
@@ -201,12 +223,15 @@ typedef enum {
 
 
 /*! \typedef pwmd_ip_version_t
+ * \brief IP protocol version for remote SSH connections.
  *
  * The value of the option \ref PWMD_OPTION_IP_VERSION which is set with \ref
  * pwmd_setopt().
  */
 typedef enum {
-    /*! Try both IPv6 and IPv4 addresses. Note that IPv6 is tried first. */
+    /*! Try both IPv4 and IPv6 addresses. Note that IPv6 is tried first and
+     * that \ref PWMD_IP_ANY only affects a hostname and not an IP address in
+     * the address specification. */
     PWMD_IP_ANY,
 
     /*! Only try IPv4. */
@@ -218,7 +243,7 @@ typedef enum {
 
 
 /*! \def PWMD_FD_READABLE
- * \hideinitializer
+ * \brief For use with \ref pwmd_get_fds().
  *
  * Set when the file descriptor is readable.
  */
@@ -226,16 +251,14 @@ typedef enum {
 
 
 /*! \def PWMD_FD_WRITABLE
- * \hideinitializer
+ * \brief For use with \ref pwmd_get_fds().
  *
  * Set when the file descriptor is writable.
  */
 #define PWMD_FD_WRITABLE	0x02
 
 
-/*!
- * For use with \ref pwmd_get_fds().
- */
+/*! For use with \ref pwmd_get_fds(). */
 typedef struct {
     /*! The file descriptor. */
     int fd;
@@ -246,6 +269,7 @@ typedef struct {
 
 
 /*! \typedef pwmd_socket_t
+ * \brief The type of socket a handle is connected to.
  *
  * For use with \ref pwmd_socket_type().
  */
@@ -259,6 +283,7 @@ typedef enum {
 
 
 /*! \typedef pwmd_pinentry_t
+ * \brief Local pinentry commands and not pwmd pinentry.
  *
  * For use with \ref pwmd_getpin().
  */
@@ -288,6 +313,7 @@ typedef enum {
 
 
 /*! \typedef pwmd_passphrase_cb_t
+ * \brief Custom passphrase retrieval function.
  *
  * The value of the option \ref PWMD_OPTION_PASSPHRASE_CB which is set with
  * \ref pwmd_setopt().
@@ -303,6 +329,7 @@ typedef gpg_error_t (*pwmd_passphrase_cb_t)(void *user, char **passphrase);
 
 
 /*! \typedef pwmd_status_cb_t
+ * \brief Process status messages.
  *
  * The value of the option \ref PWMD_OPTION_STATUS_CB which is set with \ref
  * pwmd_setopt().
@@ -316,6 +343,7 @@ typedef int (*pwmd_status_cb_t)(void *user, const char *line);
 
 
 /*! \typedef pwmd_inquire_cb_t
+ * \brief Send data to the pwmd server.
  *
  * This is a callback function that gets passed to \ref pwmd_inquire(). It is
  * used for sending data to the server for commands that need to reply to an
@@ -334,6 +362,7 @@ typedef int (*pwmd_status_cb_t)(void *user, const char *line);
  * \param[out] len The length of \a data or 0.
  *
  * \retval 0 There is more data to be sent.
+ * \retval GPG_ERR_CANCELED Cancel the current inquire.
  * \retval GPG_ERR_EOF No need to call this function again, the current
  * \a line is the last to send.
  * \retval code Any other error code which will terminate the INQUIRE.
@@ -346,9 +375,38 @@ typedef gpg_error_t (*pwmd_inquire_cb_t)(void *user, const char *cmd,
 	gpg_error_t rc, char **data, size_t *len);
 
 
-/*! \enum pwmd_option_t
+/*! \typedef pwmd_knownhost_cb_t
+ * \brief Verify a remote SSH connection.
  *
- * libpwmd options which are set with \ref pwmd_setopt().
+ * When \ref PWMD_OPTION_KNOWNHOST_CB is set and a the current connections
+ * host key was not found in the known hosts file, then this callback function
+ * can be used to confirm the addition of the new host key to the known_hosts
+ * file.
+ *
+ * \param user User data which was set with \ref PWMD_OPTION_KNOWNHOST_DATA.
+ * \param host The hostname as passed to the connecting function.
+ * \param key The raw host key. Note that this differs from the format
+ * returned from \ref pwmd_get_hostkey().
+ * \param len The host key length.
+ * \retval 0 Add the host key to the known hosts file.
+ * \retval GPG_ERR_NOT_CONFIRMED Do not add the host key and abort the
+ * connection.
+ *
+ * \note If the known hosts file cannot be modified do to filesystem
+ * restrictions when trying to add the new host key, no error is returned.
+ * Instead the host key is added to the current connections host key cache and
+ * the connection is considered verified.
+ *
+ * \see \ref ssh
+ */
+typedef gpg_error_t (*pwmd_knownhost_cb_t)(void *user, const char *host,
+	const char *key, size_t len);
+
+
+/*! \enum pwmd_option_t
+ * \brief libpwmd options.
+ *
+ * Options are set with \ref pwmd_setopt().
  *
  * \filepath
  */
@@ -453,6 +511,57 @@ typedef enum {
      * \pre_conn_req
      */
     PWMD_OPTION_IP_VERSION,
+
+    /*! A function to confirm an unknown host hash which wasn't found in the
+     * known hosts file.
+     *
+     * \see \ref ssh
+     */
+    PWMD_OPTION_KNOWNHOST_CB,
+
+    /*! User supplied data which is passed to the known host function.
+     *
+     * \see \ref ssh
+     */
+    PWMD_OPTION_KNOWNHOST_DATA,
+
+    /*! When the total number of bytes to be sent via an INQUIRE is known,
+     * this should be set so XFER status messages can be parsed correctly.
+     * When not known or unset, 0 is used as the total argument to the XFER
+     * status message. This option should be set before each call to
+     * \ref pwmd_inquire().
+     *
+     * \note During the INQUIRE, PWMD_OPTION_STATUS_CB is called after every
+     * ASSUAN_LINELENGTH bytes have been successfully transferred.
+     *
+     * \note This is a libpwmd feature. pwmd itself does not send XFER status
+     * messages during an INQUIRE. Status messages can be parsed only when
+     * PWMD_OPTION_STATUS_CB is set.
+     */
+    PWMD_OPTION_INQUIRE_TOTAL,
+
+    /*! When set to 1, lock the file mutex after opening a file as if the LOCK
+     * command had been sent.
+     */
+    PWMD_OPTION_LOCK_ON_OPEN,
+
+    /*! A long integer specifying the number of iterations to encrypt with.
+     * When -1 then the next save operation will use the current iteration
+     * setting and no command will be sent to alter it.
+     */
+    PWMD_OPTION_ITERATIONS,
+
+    /*! A string specifying the cipher to use to encrypt with. See the pwmd(1)
+     * manual page for available ciphers.
+     */
+    PWMD_OPTION_CIPHER,
+
+    /*! When 1, tell pwmd that the passphrase is Base64 encoded. pwmd will
+     * decode the passphrase before encryption and decryption. */
+    PWMD_OPTION_BASE64,
+
+    /*! Use ssh-agent to retrieve private key to authenticate. */
+    PWMD_OPTION_SSH_AGENT,
 } pwmd_option_t;
 
 
@@ -464,7 +573,7 @@ typedef enum {
  *
  * \return 0 on success or an error code.
  */
-gpg_error_t pwmd_init(void);
+LIBPWMD_API gpg_error_t pwmd_init(void);
 
 
 /*! \brief Creates a new handle.
@@ -476,8 +585,7 @@ gpg_error_t pwmd_init(void);
  *
  * \return a new handle or NULL if there was not enough memory.
  */
-pwm_t *pwmd_new(const char *name)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API pwm_t *pwmd_new(const char *name);
 
 
 /*! \brief Connect to a local pwmd server.
@@ -491,8 +599,7 @@ pwm_t *pwmd_new(const char *name)
  * \filepath
  * \see pwmd_ssh_connect(), pwmd_ssh_connect_async(), pwmd_disconnect()
  */
-gpg_error_t pwmd_connect(pwm_t *pwm, const char *path)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_connect(pwm_t *pwm, const char *path);
 
 
 /*! \brief Establish a remote connection to a pwmd server.
@@ -500,23 +607,27 @@ gpg_error_t pwmd_connect(pwm_t *pwm, const char *path)
  * Connects to a pwmd server over an SSH channel.
  *
  * \param pwm A handle.
- * \param host The hostname to connect to.
+ * \param host The hostname to connect to or NULL to resume a connection
+ * previously started with \ref pwmd_get_hostkey().
  * \param port The port or -1 for the default of 22.
  * \param identity The SSH identity file to use for authentication. This
  * should specify the private key. The public key is assumed to be \a
- * identity.pub.
+ * identity.pub. This parameter is not used if \ref PWMD_OPTION_SSH_AGENT is
+ * set.
  * \param user The username on the SSH server to login as. If NULL then
  * invoking username will be used.
- * \param known_hosts A file containing the public SSH server key hash in SHA1
- * format which may be obtained with \ref pwmd_get_hostkey().
+ * \param known_hosts An OpenSSH known hosts formatted file containing public
+ * SSH server hashes which may be obtained with \ref pwmd_get_hostkey() or via
+ * \ref pwmd_knownhost_cb_t during a connection. If NULL, the default of
+ * \a "~/.ssh/known_hosts" will be used.
  * \return 0 on success or an error code.
  * \filepath
  * \see pwmd_ssh_connect_async(), \ref PWMD_OPTION_IP_VERSION,
  * pwmd_disconnect(), \ref ssh
  */
-gpg_error_t pwmd_ssh_connect(pwm_t *pwm, const char *host, int port,
-	const char *identity, const char *user, const char *known_hosts)
-    __attribute__ ((warn_unused_result)); 
+LIBPWMD_API gpg_error_t pwmd_ssh_connect(pwm_t *pwm, const char *host,
+	int port, const char *identity, const char *user, const char
+	*known_hosts);
 
 
 /*! \brief Establish a remote connection to a pwmd server (asynchronously).
@@ -527,23 +638,27 @@ gpg_error_t pwmd_ssh_connect(pwm_t *pwm, const char *host, int port,
  * \process
  *
  * \param pwm A handle.
- * \param host The hostname to connect to.
+ * \param host The hostname to connect to or NULL to resume a connection
+ * previously started with \ref pwmd_get_hostkey().
  * \param port The port or -1 for the default of 22.
  * \param identity The SSH identity file to use for authentication. This
  * should specify the private key. The public key is assumed to be \a
- * identity.pub.
+ * identity.pub. This parameter is not used if \ref PWMD_OPTION_SSH_AGENT is
+ * set.
  * \param user The username on the SSH server to login as. If NULL, the
  * invoking username will be used.
- * \param known_hosts A file containing the public SSH server key hash in SHA1
- * format which may be obtained with \ref pwmd_get_hostkey().
+ * \param known_hosts An OpenSSH known hosts formatted file containing public
+ * SSH server hashes which may be obtained with \ref pwmd_get_hostkey() or via
+ * \ref pwmd_knownhost_cb_t during a connection. If NULL, the default of
+ * \a "~/.ssh/known_hosts" will be used.
  * \return 0 on success or an error code.
  * \filepath
  * \see pwmd_process(), \ref PWMD_OPTION_IP_VERSION, pwmd_disconnect(),
  * \ref ssh
  */
-gpg_error_t pwmd_ssh_connect_async(pwm_t *pwm, const char *host, int port,
-	const char *identity, const char *user, const char *known_hosts)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_ssh_connect_async(pwm_t *pwm, const char *host,
+	int port, const char *identity, const char *user, const char
+	*known_hosts);
 
 
 /*! \brief Establish a connection by parsing a URL.
@@ -556,7 +671,7 @@ gpg_error_t pwmd_ssh_connect_async(pwm_t *pwm, const char *host, int port,
  *
  * or
  *
- * ssh[46]://[username@]hostname[:port],identity,known_hosts
+ * ssh[46]://[username@]hostname[:port][,identity,known_hosts]
  * \endcode
  *
  * The parameters in square brackets are optional and if not specified then
@@ -569,8 +684,7 @@ gpg_error_t pwmd_ssh_connect_async(pwm_t *pwm, const char *host, int port,
  * \return 0 on success or an error code.
  * \see \ref pwmd_socket_type(), pwmd_disconnect()
  */
-gpg_error_t pwmd_connect_url(pwm_t *pwm, const char *url)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_connect_url(pwm_t *pwm, const char *url);
 
 
 /*! \brief Establish a connection by parsing a URL (asynchronously).
@@ -583,7 +697,7 @@ gpg_error_t pwmd_connect_url(pwm_t *pwm, const char *url)
  *
  * or
  *
- * ssh[46]://[username@]hostname[:port],identity,known_hosts
+ * ssh[46]://[username@]hostname[:port][,identity,known_hosts]
  * \endcode
  *
  * The parameters in square brackets are optional and if not specified then
@@ -598,27 +712,33 @@ gpg_error_t pwmd_connect_url(pwm_t *pwm, const char *url)
  * \return 0 on success or an error code.
  * \see \ref pwmd_socket_type(), pwmd_disconnect()
  */
-gpg_error_t pwmd_connect_url_async(pwm_t *pwm, const char *url)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_connect_url_async(pwm_t *pwm, const char *url);
 
 
-/*! \brief Retrieve a remote SSH host key.
+/*! \brief Retrieve a remote SSH public host key.
  *
  * This key is needed for host verification of the remote pwmd server. You
  * should be sure that the remote host is really the host that your wanting to
  * connect to and not subject to a man-in-the-middle attack.
  *
  * \param pwm A handle.
- * \param host The hostname to connect to.
+ * \param host The hostname or IP to connect to.
  * \param port The port or -1 for the default of 22.
- * \param[out] result The SHA1 sum of the server host key which must be freed
- * with \ref pwmd_free().
+ * \param[out] result An OpenSSH known hosts formatted string containing the
+ * servers public key which should be freed with \ref pwmd_free(). If the
+ * \a host was a hostname then two newline separated known host entries will
+ * be returned; one for the hostname and one for the resolved IP address. The
+ * IP known host entry will always be the second in the string.
  * \return 0 on success or an error code.
- * \see pwmd_get_hostkey_async(), \ref ssh
+ *
+ * \version 6.0.3
+ * The connection is kept open but not verified after returning. It can be
+ * resumed from one of the SSH connection functions.
+ *
+ * \see pwmd_get_hostkey_async(), pwmd_ssh_connect(), \ref ssh
  */
-gpg_error_t pwmd_get_hostkey(pwm_t *pwm, const char *host, int port,
-	char **result)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_get_hostkey(pwm_t *pwm, const char *host, int port,
+	char **result);
 
 
 /*! \brief Retrieve a remote SSH host key (asynchronously).
@@ -630,13 +750,21 @@ gpg_error_t pwmd_get_hostkey(pwm_t *pwm, const char *host, int port,
  * \process
  *
  * \param pwm A handle.
- * \param host The hostname to connect to.
+ * \param host The hostname or IP to connect to.
  * \param port The port or -1 for the default of 22.
- * \return 0 on success or an error code.
- * \see pwmd_get_hostkey(), \ref pwmd_process(), \ref ssh
+ * \return 0 on success or an error code. The result is obtained from \ref
+ * pwmd_process() should be freed with \ref pwmd_free(). It has the same
+ * format as the result from \ref pwmd_get_hostkey().
+ *
+ * \version 6.0.3
+ * The connection is kept open but not verified after returning. It can be
+ * resumed from one of the SSH connection functions.
+ *
+ * \see pwmd_get_hostkey(), pwmd_ssh_connect_async(), \ref pwmd_process(),
+ * \ref ssh
  */
-gpg_error_t pwmd_get_hostkey_async(pwm_t *pwm, const char *host, int port)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_get_hostkey_async(pwm_t *pwm, const char *host,
+	int port);
 
 
 /*! \brief Get the associated file descriptor(s) for a handle.
@@ -661,8 +789,7 @@ gpg_error_t pwmd_get_hostkey_async(pwm_t *pwm, const char *host, int port)
  * specified in \a n_fds. \a fds and \a n_fds are still usable though.
  * \see pwmd_process()
  */
-gpg_error_t pwmd_get_fds(pwm_t *pwm, pwmd_fd_t *fds, int *n_fds)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_get_fds(pwm_t *pwm, pwmd_fd_t *fds, int *n_fds);
 
 
 /*! \brief Check for a unparsed buffered line.
@@ -675,8 +802,7 @@ gpg_error_t pwmd_get_fds(pwm_t *pwm, pwmd_fd_t *fds, int *n_fds)
  * \retval GPG_ERR_NO_DATA if there is no pending line.
  * \see pwmd_process()
  */
-gpg_error_t pwmd_pending_line(pwm_t *pwm)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_pending_line(pwm_t *pwm);
 
 
 /*! \brief Set handle options.
@@ -687,8 +813,7 @@ gpg_error_t pwmd_pending_line(pwm_t *pwm)
  * \param opt The option and following value.
  * \return 0 on success or an error code.
  */
-gpg_error_t pwmd_setopt(pwm_t *pwm, pwmd_option_t opt, ...)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_setopt(pwm_t *pwm, pwmd_option_t opt, ...);
 
 
 /*! \brief Launch a local pinentry.
@@ -713,9 +838,8 @@ gpg_error_t pwmd_setopt(pwm_t *pwm, pwmd_option_t opt, ...)
  *
  * \return 0 on success or an error.
  */
-gpg_error_t pwmd_getpin(pwm_t *pwm, const char *filename, char **result,
-	pwmd_pinentry_t which)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_getpin(pwm_t *pwm, const char *filename,
+	char **result, pwmd_pinentry_t which);
 
 
 /*! \brief Open a file on the pwmd server.
@@ -728,8 +852,34 @@ gpg_error_t pwmd_getpin(pwm_t *pwm, const char *filename, char **result,
  * \return 0 on success or an error code.
  * \see \ref pinentry
  */
-gpg_error_t pwmd_open(pwm_t *pwm, const char *filename)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_open(pwm_t *pwm, const char *filename);
+
+
+/*! \brief Open a file on the pwmd server by using a server inquire.
+ *
+ * This is a convenience function to the OPEN command using \ref
+ * pwmd_inquire() in that it passes the OPEN options that were set with \ref
+ * pwmd_setopt() for you. It uses the specified callback function to retreive
+ * the passphrase.
+ *
+ * \note Only the passphrase should be sent when using this function and not
+ * any filename argument. That argument will be sent automatically when using
+ * this function.
+ *
+ * \note Pinentry is disabled when using this function.
+ *
+ * \param pwm A handle.
+ * \param filename The filename to open. The \a filename is not a full path
+ * but the data file only.
+ * \param func A callback function of type \ref pwmd_inquire_cb_t which sets
+ * the passphrase to be sent. This acts as a normal \ref pwmd_inquire() would.
+ * \param user A user data pointer passed to the callback function \a func.
+ * \return 0 on success or an error code.
+ * \see \ref PWMD_OPTION_BASE64
+ */
+LIBPWMD_API gpg_error_t pwmd_open_inquire(pwm_t *pwm, const char *filename,
+	pwmd_inquire_cb_t func, void *user);
+
 
 /*! \brief Open a file on the pwmd server using a local pinentry.
  *
@@ -748,8 +898,7 @@ gpg_error_t pwmd_open(pwm_t *pwm, const char *filename)
  * \retval GPG_ERR_PIN_BLOCKED Another handle is using the local pinentry.
  * \see \ref pinentry
  */
-gpg_error_t pwmd_open2(pwm_t *pwm, const char *filename)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_open2(pwm_t *pwm, const char *filename);
 
 
 /*! \brief Open a file on the pwmd server asynchronously (fork method).
@@ -768,8 +917,7 @@ gpg_error_t pwmd_open2(pwm_t *pwm, const char *filename)
  * \return 0 on success or an error code.
  * \see pwmd_process(), \ref pinentry
  */
-gpg_error_t pwmd_open_async2(pwm_t *pwm, const char *filename)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_open_async2(pwm_t *pwm, const char *filename);
 
 
 /*! \brief Open a file on the pwmd server (asynchronously).
@@ -786,8 +934,7 @@ gpg_error_t pwmd_open_async2(pwm_t *pwm, const char *filename)
  * \return 0 on success or an error code.
  * \see pwmd_process(), \ref pinentry
  */
-gpg_error_t pwmd_open_async(pwm_t *pwm, const char *filename)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_open_async(pwm_t *pwm, const char *filename);
 
 
 /*! \brief Process an asynchronous function.
@@ -811,8 +958,8 @@ gpg_error_t pwmd_open_async(pwm_t *pwm, const char *filename)
  * be called again.
  * \see pwmd_get_fds(), pwmd_pending_line()
  */
-pwmd_async_t pwmd_process(pwm_t *pwm, gpg_error_t *rc, char **result)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API pwmd_async_t pwmd_process(pwm_t *pwm, gpg_error_t *rc,
+	char **result);
 
 
 /*! \brief Save a file on the pwmd server.
@@ -823,8 +970,27 @@ pwmd_async_t pwmd_process(pwm_t *pwm, gpg_error_t *rc, char **result)
  * \return 0 on success or an error code.
  * \see \ref pinentry
  */
-gpg_error_t pwmd_save(pwm_t *pwm)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_save(pwm_t *pwm);
+
+
+/*! \brief Save a file on the pwmd server by using a server inquire.
+ *
+ * This is a convenience function to the SAVE command using \ref
+ * pwmd_inquire() in that it passes the SAVE options that were set with \ref
+ * pwmd_setopt() for you. It uses the specified callback function to retreive
+ * the passphrase.
+ *
+ * \note Pinentry is disabled when using this function.
+ *
+ * \param pwm A handle.
+ * \param func A callback function of type \ref pwmd_inquire_cb_t which sets
+ * the passphrase to be sent. This acts as a normal \ref pwmd_inquire() would.
+ * \param user A user data pointer passed to the callback function \a func.
+ * \return 0 on success or an error code.
+ * \see \ref PWMD_OPTION_BASE64
+ */
+LIBPWMD_API gpg_error_t pwmd_save_inquire(pwm_t *pwm, pwmd_inquire_cb_t func,
+	void *user);
 
 
 /*! \brief Save a file on the pwmd server using the local pinentry.
@@ -836,8 +1002,7 @@ gpg_error_t pwmd_save(pwm_t *pwm)
  * \return 0 on success or an error code.
  * \see \ref pinentry
  */
-gpg_error_t pwmd_save2(pwm_t *pwm)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_save2(pwm_t *pwm);
 
 
 /*! \brief Save a file on the pwmd server asynchronously (fork method).
@@ -852,8 +1017,7 @@ gpg_error_t pwmd_save2(pwm_t *pwm)
  * \return 0 on success or an error code.
  * \see pwmd_process(), \ref pinentry
  */
-gpg_error_t pwmd_save_async2(pwm_t *pwm)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_save_async2(pwm_t *pwm);
 
 
 /*! \brief Save changes to a file on the pwmd server (asynchronously).
@@ -868,8 +1032,7 @@ gpg_error_t pwmd_save_async2(pwm_t *pwm)
  * \return 0 on success or an error code.
  * \see pwmd_process(), \ref pinentry
  */
-gpg_error_t pwmd_save_async(pwm_t *pwm)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_save_async(pwm_t *pwm);
 
 
 /*! \brief Send a command to the pwmd server.
@@ -888,8 +1051,8 @@ gpg_error_t pwmd_save_async(pwm_t *pwm)
  *
  * \note Not all commands return a result.
  */
-gpg_error_t pwmd_command(pwm_t *pwm, char **result, const char *cmd, ...)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_command(pwm_t *pwm, char **result,
+	const char *cmd, ...);
 
 
 /*! \brief Send a command to the pwmd server.
@@ -905,9 +1068,8 @@ gpg_error_t pwmd_command(pwm_t *pwm, char **result, const char *cmd, ...)
  *
  * \note Not all commands return a result.
  */
-gpg_error_t pwmd_command_ap(pwm_t *pwm, char **result, const char *cmd,
-	va_list ap)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_command_ap(pwm_t *pwm, char **result,
+	const char *cmd, va_list ap);
 
 
 /*! \brief Send data to a pwmd server.
@@ -917,7 +1079,7 @@ gpg_error_t pwmd_command_ap(pwm_t *pwm, char **result, const char *cmd,
  * pwmd commands.
  *
  * \param pwm A handle.
- * \param cmd The command (without arguments) to send that uses an INQUIRE.
+ * \param cmd The \ref pwmd(1) command to send including any options.
  * \param func A callback function of type \ref pwmd_inquire_cb_t which sets
  * the data to be sent.
  * \param user A user data pointer passed to the callback function \a func.
@@ -925,9 +1087,8 @@ gpg_error_t pwmd_command_ap(pwm_t *pwm, char **result, const char *cmd,
  *
  * \see pwmd_inquire_cb_t
  */
-gpg_error_t pwmd_inquire(pwm_t *pwm, const char *cmd, pwmd_inquire_cb_t func,
-	void *user)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_inquire(pwm_t *pwm, const char *cmd,
+	pwmd_inquire_cb_t func, void *user);
 
 
 /*! \brief Close a connection to the pwmd server.
@@ -939,8 +1100,7 @@ gpg_error_t pwmd_inquire(pwm_t *pwm, const char *cmd, pwmd_inquire_cb_t func,
  * \return 0 on success or an error code.
  * \see \ref pwmd_close()
  */
-gpg_error_t pwmd_disconnect(pwm_t *pwm)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_disconnect(pwm_t *pwm);
 
 
 /*! \brief Close a handle.
@@ -952,7 +1112,7 @@ gpg_error_t pwmd_disconnect(pwm_t *pwm)
  * \return Nothing.
  * \see \ref pwmd_disconnect(), \ref pwmd_new()
  */
-void pwmd_close(pwm_t *pwm);
+LIBPWMD_API void pwmd_close(pwm_t *pwm);
 
 
 /*! \brief The type of connection a handle has.
@@ -964,8 +1124,7 @@ void pwmd_close(pwm_t *pwm);
  * \return 0 on success or an error code.
  * \see pwmd_connect_url()
  */
-gpg_error_t pwmd_socket_type(pwm_t *pwm, pwmd_socket_t *type)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API gpg_error_t pwmd_socket_type(pwm_t *pwm, pwmd_socket_t *type);
 
 
 /*! \brief Free a previously allocated pointer.
@@ -981,7 +1140,7 @@ gpg_error_t pwmd_socket_type(pwm_t *pwm, pwmd_socket_t *type)
  * \see pwmd_malloc(), pwmd_calloc(), pwmd_realloc(), pwmd_strdup(),
  * pwmd_process(), pwmd_command()
  */
-void pwmd_free(void *ptr);
+LIBPWMD_API void pwmd_free(void *ptr);
 
 
 /*! \brief A wrapper around malloc.
@@ -992,8 +1151,7 @@ void pwmd_free(void *ptr);
  * \return A newly allocated pointer or NULL if there wasn't enough memory.
  * \see malloc(3), pwmd_free()
  */
-void *pwmd_malloc(size_t size)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API void *pwmd_malloc(size_t size);
 
 
 /*! \brief A wrapper around calloc().
@@ -1005,8 +1163,7 @@ void *pwmd_malloc(size_t size)
  * \return A newly allocated pointer or NULL if there wasn't enough memory.
  * \see calloc(3), pwmd_free()
  */
-void *pwmd_calloc(size_t nmemb, size_t size)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API void *pwmd_calloc(size_t nmemb, size_t size);
 
 
 /*! \brief A wrapper around realloc().
@@ -1021,8 +1178,7 @@ void *pwmd_calloc(size_t nmemb, size_t size)
  * \return A newly allocated pointer or NULL if there wasn't enough memory.
  * \see realloc(3), pwmd_free()
  */
-void *pwmd_realloc(void *ptr, size_t size)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API void *pwmd_realloc(void *ptr, size_t size);
 
 
 /*! \brief A wrapper around strdup().
@@ -1034,8 +1190,8 @@ void *pwmd_realloc(void *ptr, size_t size)
  * enough memory.
  * \see strdup(3), pwmd_free()
  */
-char *pwmd_strdup(const char *str)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API char *pwmd_strdup(const char *str);
+
 
 /*! \brief Duplicate a formatted string.
  *
@@ -1046,40 +1202,48 @@ char *pwmd_strdup(const char *str)
  * enough memory.
  * \see pwmd_free()
  */
-char *pwmd_strdup_printf(const char *fmt, ...)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API char *pwmd_strdup_printf(const char *fmt, ...);
 
 
 /*! \def EPWMD_NO_FILE
- *  \hideinitializer
+ * \hideinitializer
+ * \brief No data file has been opened.
  *
- * A command required an open file but no file has yet been opened.
+ * Some commands don't require an open data file but most do.
  */
 #define EPWMD_NO_FILE		GPG_ERR_USER_1
 
 
 /*! \def EPWMD_LIBXML_ERROR
- *  \hideinitializer
+ * \hideinitializer
+ * \brief libxml2 error.
  *
- * An XML parse or other libxml2 error occurred.
+ * This can be a memory allocation error or a parse error. The details of the
+ * error cannot be obtained with libpwmd. You'd have to connect to the pwmd
+ * socket and do the command directly to get the actual error.
  */
 #define EPWMD_LIBXML_ERROR	GPG_ERR_USER_2
 
 
 /*! \def EPWMD_FILE_MODIFIED
- *  \hideinitializer
+ * \hideinitializer
+ * \brief The data file has been modified.
  *
- * The data file was modified either externally or by another client while
- * trying to process a command.
+ * Rather than process the next command this error is returned to prevent
+ * overwriting new data which may have been saved by another client.
  */
 #define EPWMD_FILE_MODIFIED	GPG_ERR_USER_3
 
 
 /*! \def EPWMD_MAX
- *  \hideinitializer
+ * \hideinitializer
  * \if cond1
  * \internal
  * \endif
+ * \brief libgpg-error error code offset.
+ *
+ * If you use your own libgpg-error codes then this should be the base of
+ * them.
  */
 #define EPWMD_MAX		GPG_ERR_USER_4
 
@@ -1090,8 +1254,7 @@ char *pwmd_strdup_printf(const char *fmt, ...)
  * \return A character description of the error code.
  * \see pwmd_strerror_r()
  */
-const char *pwmd_strerror(gpg_error_t code)
-    __attribute__ ((warn_unused_result));
+LIBPWMD_API const char *pwmd_strerror(gpg_error_t code);
 
 
 /*! \brief Return a description of an error code (thread-safe).
@@ -1106,7 +1269,7 @@ const char *pwmd_strerror(gpg_error_t code)
  * \retval ERANGE \a size was not large enough to hold the entire description
  * and \a buf is set to the truncated error string.
  */
-int pwmd_strerror_r(gpg_error_t code, char *buf, size_t size);
+LIBPWMD_API int pwmd_strerror_r(gpg_error_t code, char *buf, size_t size);
 
 #ifdef __cplusplus
 }
