@@ -1,6 +1,6 @@
 /* vim:tw=78:ts=8:sw=4:set ft=c:  */
 /*
-    Copyright (C) 2007-2009 Ben Kibbey <bjk@luxsci.net>
+    Copyright (C) 2007-2010 Ben Kibbey <bjk@luxsci.net>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -16,6 +16,10 @@
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02110-1301  USA
 */
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -24,14 +28,11 @@
 #include <string.h>
 #include <libpwmd.h>
 #include <assuan.h>
-#ifdef DEBUG
 #include <sys/select.h>
-#endif
 #include <fcntl.h>
-
-#ifdef HAVE_CONFIG_H
-#include <config.h>
-#endif
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <libgen.h>
 
 #ifdef HAVE_LOCALE_H
 #include <locale.h>
@@ -45,13 +46,36 @@
 #include "getopt_long.h"
 #endif
 
+#ifdef HAVE_LIBREADLINE
+#  if defined(HAVE_READLINE_READLINE_H)
+#    include <readline/readline.h>
+#  elif defined(HAVE_READLINE_H)
+#    include <readline.h>
+#  endif /* !defined(HAVE_READLINE_H) */
+#endif /* HAVE_LIBREADLINE */
+
+#ifdef HAVE_READLINE_HISTORY
+#  if defined(HAVE_READLINE_HISTORY_H)
+#    include <readline/history.h>
+#  elif defined(HAVE_HISTORY_H)
+#    include <history.h>
+#  endif
+#endif /* HAVE_READLINE_HISTORY */
+
 #include "gettext.h"
 #define N_(msgid)	gettext(msgid)
 
 #include "mem.h"
 
 #define DEFAULT_PORT	22
-pwm_t *pwm;
+
+static pwm_t *pwm;
+
+struct inquire_s {
+    int fd;
+    char *line;
+    size_t len;
+};
 
 static void show_error(gpg_error_t error)
 {
@@ -73,17 +97,20 @@ static void usage(const char *pn, int status)
 	    "   --host, -h <hostname>\n"
 	    "       connect to the specified hostname\n"
 	    "\n"
-	    "   --port\n"
+	    "   --port <N>\n"
 	    "       alterate port (22)\n"
 	    "\n"
-	    "   --user\n"
+	    "   --user <username>\n"
 	    "       SSH username (default is the invoking user)\n"
 	    "\n"
+	    "   --use-agent\n"
+	    "       use the SSH agent for authentication\n"
+	    "\n"
 	    "   --identity, -i <filename>\n"
-	    "       SSH identity file\n"
+	    "       SSH identity file (if not using the SSH agent)\n"
 	    "\n"
 	    "   --known-hosts, -k <filename>\n"
-	    "       known host's file (for server validation)\n"
+	    "       known hosts file (~/.ssh/known_hosts)\n"
 	    "\n"
 	    "   --get-hostkey, -g\n"
 	    "       retrieve the remote SSH host key and exit\n"
@@ -96,7 +123,7 @@ static void usage(const char *pn, int status)
 	    "\n"
 #endif
             "   --url <string>\n"
-	    "       a url string to parse\n"
+	    "       a url string to parse (see below)\n"
 	    "\n"
 	    "   --no-status\n"
 	    "       disable showing of status messages from the server\n"
@@ -110,6 +137,12 @@ static void usage(const char *pn, int status)
 	    "   --passphrase, -P <string>\n"
 	    "       passphrase to use (disables pinentry use)\n"
 	    "\n"
+	    "   --key-file <filename>\n"
+	    "       obtain the passphrase from the specified filename\n"
+	    "\n"
+	    "   --base64\n"
+	    "       the passphrase is base64 encoded\n"
+	    "\n"
 	    "   --timeout <seconds>\n"
 	    "       pinentry timeout\n"
 	    "\n"
@@ -123,10 +156,10 @@ static void usage(const char *pn, int status)
 	    "       tty that pinentry will use\n"
 	    "\n"
 	    "   --ttytype, -t <string>\n"
-	    "       pinentry terminal type (default is TERM)\n"
+	    "       pinentry terminal type (default is $TERM)\n"
 	    "\n"
 	    "   --display, -d\n"
-	    "       pinentry display (default is DISPLAY)\n"
+	    "       pinentry display (default is $DISPLAY)\n"
 	    "\n"
 	    "   --lc-ctype <string>\n"
 	    "       locale setting for pinentry\n"
@@ -134,24 +167,36 @@ static void usage(const char *pn, int status)
 	    "   --lc-messages <string>\n"
 	    "       locale setting for pinentry\n"
 	    "\n"
-
 	    "   --local-pinentry\n"
 	    "       force using a local pinentry\n"
 	    "\n"
+	    "   --interactive\n"
+	    "       use a shell like interface to pwmd (allows more than one command)\n"
+	    "\n"
 	    "   --output-fd <FD>\n"
 	    "       redirect command output to the specified file descriptor\n"
 	    "\n"
+	    "   --inquire <COMMAND>\n"
+	    "       the specified command (with any options) uses a server inquire while\n"
+	    "       command data is read via the inquire file descriptor (stdin)\n"
+	    "\n"
 	    "   --inquire-fd <FD>\n"
-	    "       read inquire data from the specified file descriptor\n"
+	    "       read inquire data from the specified file descriptor (stdin)\n"
+	    "\n"
+	    "   --inquire-line, -L <STRING>\n"
+	    "       the initial line to send (i.e., element path) before the inquire data\n"
+	    "\n"
+	    "   --cipher <string>\n"
+	    "       the cipher to use when saving (see pwmd(1))\n"
 	    "\n"
 	    "   --save, -S\n"
 	    "       send the SAVE command before exiting\n"
 	    "\n"
 	    "   --force-save\n"
-	    "       like --save, but ask for a passphrase\n"
+	    "       like --save but always ask for a passphrase\n"
 	    "\n"
 	    "   --iterations, -I <N>\n"
-	    "       encrypt with the specified number of iterations when saving\n"
+	    "       encrypt with the specified number of iterations\n"
 	    "\n"
 	    "   --version\n"
 	    "   --help\n"));
@@ -160,60 +205,42 @@ static void usage(const char *pn, int status)
 		"A url string (specified with --url) may be in the form of:\n"
 		"    file://[path/to/socket]\n"
 #ifdef WITH_TCP
-		"    ssh[46]://[username@]hostname[:port],identity,known_hosts\n"
+		"    ssh[46]://[username@]hostname[:port],identity[,known_hosts]\n"
+		"    or\n"
+		"    ssh[46]://[username@]hostname[:port][,,known_hosts] --use-agent\n"
 #endif
 		));
     exit(status);
 }
 
-struct inquire_s {
-    FILE *fp;
-    char *data;
-};
-
-static gpg_error_t do_inquire(void *data, const char *keyword, gpg_error_t rc,
-	char **result, size_t *result_len)
+static gpg_error_t inquire_cb(void *user, const char *cmd, gpg_error_t rc,
+	char **data, size_t *len)
 {
-    int c;
-    static char buf[ASSUAN_LINELENGTH];
-    char *p;
-    size_t len = 0;
-    struct inquire_s *inq = (struct inquire_s *)data;
+    struct inquire_s *inq = user;
 
-    if (rc) {
-	memset(buf, 0, sizeof(buf));
+    *data = NULL;
+    *len = 0;
+
+    if (rc)
 	return rc;
+
+    /* The first part of the command data. */
+    if (inq->len) {
+	*data = inq->line;
+	*len = inq->len;
+	inq->len = 0;
+	return 0;
     }
 
-    buf[0] = 0;
-    p = buf;
+    *len = read(inq->fd, inq->line, ASSUAN_LINELENGTH);
 
-    if (inq->data) {
-	snprintf(buf, sizeof(buf), "%s", inq->data);
-	pwmd_free(inq->data);
-	inq->data = NULL;
-	len = strlen(buf);
-	p = buf + len;
-    }
+    if (*len == -1)
+	return gpg_error_from_syserror();
 
-    while ((c = fgetc(inq->fp)) != EOF) {
-	if (len == sizeof(buf)) {
-	    ungetc(c, inq->fp);
-	    break;
-	}
+    if (*len)
+	*data = inq->line;
 
-	*p++ = c;
-	len++;
-    }
-
-    if (!buf[0]) {
-	memset(buf, 0, sizeof(buf));
-	return GPG_ERR_EOF;
-    }
-
-    *result = buf;
-    *result_len = len;
-    return 0;
+    return *len ? 0 : GPG_ERR_EOF;
 }
 
 static int status_msg_cb(void *data, const char *line)
@@ -222,7 +249,7 @@ static int status_msg_cb(void *data, const char *line)
     return 0;
 }
 
-static gpg_error_t process_cmd(pwm_t *pwm, char **result, int input)
+static gpg_error_t process_cmd(pwm_t *pwm, char **result, int input, int once)
 {
     gpg_error_t rc;
     pwmd_async_t s;
@@ -232,6 +259,7 @@ static gpg_error_t process_cmd(pwm_t *pwm, char **result, int input)
 	fd_set rfds;
 	int nfds = 5;
 	pwmd_fd_t pfds[nfds];
+	struct timeval tv = {0, 50000};
 
 	FD_ZERO(&rfds);
 	rc = pwmd_get_fds(pwm, pfds, &nfds);
@@ -239,8 +267,10 @@ static gpg_error_t process_cmd(pwm_t *pwm, char **result, int input)
 	if (rc)
 	    return rc;
 
-	if (!nfds)
-	    return 0;
+	if (!nfds) {
+	    s = pwmd_process(pwm, &rc, result);
+	    break;
+	}
 
 	for (i = 0, n = 0; i < nfds; i++) {
 	    FD_SET(pfds[i].fd, &rfds);
@@ -250,7 +280,7 @@ static gpg_error_t process_cmd(pwm_t *pwm, char **result, int input)
 	if (input)
 	    FD_SET(STDIN_FILENO, &rfds);
 
-	nfds = select(n+1, &rfds, NULL, NULL, NULL);
+	nfds = select(n+1, &rfds, NULL, NULL, &tv);
 
 	if (nfds == -1) {
 	    rc = gpg_error_from_errno(errno);
@@ -261,15 +291,213 @@ static gpg_error_t process_cmd(pwm_t *pwm, char **result, int input)
 	    return 0;
 
 	s = pwmd_process(pwm, &rc, result);
+
+	if (once)
+	    break;
     } while (s == ASYNC_PROCESS);
 
     return rc;
 }
 
+#ifdef WITH_TCP
+static gpg_error_t knownhost_cb(void *data, const char *host, const char *key,
+	size_t len)
+{
+    gpg_error_t rc;
+    char *buf = pwmd_strdup_printf(N_("Password Manager Daemon: %s\n\nWhile attempting an SSH connection to %s there was a problem verifying it's hostkey against the known and trusted hosts file because it's hostkey was not found.\n\nWould you like to treat this connection as trusted for this and future connections by adding %s's hostkey to the known hosts file?"), (char *)data, host, host);
+
+    rc = pwmd_setopt(pwm, PWMD_OPTION_PINENTRY_TITLE, buf);
+    pwmd_free(buf);
+
+    if (rc)
+	return rc;
+
+    return pwmd_getpin(pwm, NULL, NULL, PWMD_PINENTRY_CONFIRM);
+}
+
+static int is_remote_url(const char *str)
+{
+    if (strstr(str, "file://") || strstr(str, "local://"))
+	return 0;
+
+    return 1;
+}
+#endif
+
+static char *escape(const char *str)
+{
+    const char *p;
+    char *buf = pwmd_malloc(ASSUAN_LINELENGTH+1), *b = buf;
+    size_t len = 0;
+
+    for (p = str; *p; p++, len++) {
+	if (len == ASSUAN_LINELENGTH)
+	    break;
+
+	if (*p == '\\') {
+	    switch (*++p) {
+		case 't':
+		    *b++ = '\t';
+		    break;
+		case 'n':
+		    *b++ = '\n';
+		    break;
+		case 'v':
+		    *b++ = '\v';
+		    break;
+		case 'b':
+		    *b++ = '\b';
+		    break;
+		case 'f':
+		    *b++ = '\f';
+		    break;
+		case 'r':
+		    *b++ = '\r';
+		    break;
+		default:
+		    *b++ = *p;
+		    break;
+	    }
+
+	    if (!*p)
+		break;
+
+	    continue;
+	}
+
+	*b++ = *p;
+    }
+
+    *b = 0;
+    return buf;
+}
+
+static gpg_error_t send_inquire(int fd, const char *command, const char *line)
+{
+    struct inquire_s inq;
+    struct stat st;
+    gpg_error_t rc;
+
+    if (fstat(fd, &st) == -1)
+	return gpg_error_from_syserror();
+
+    memset(&inq, 0, sizeof(inq));
+    inq.fd = fd;
+    inq.line = pwmd_calloc(1, ASSUAN_LINELENGTH);
+    inq.len = 0;
+
+    if (!inq.line)
+	return GPG_ERR_ENOMEM;
+
+    if (line) {
+	char *s = escape(line);
+
+	strncpy(inq.line, s, ASSUAN_LINELENGTH-1);
+	inq.len = strlen(s);
+	pwmd_free(s);
+    }
+
+    rc = pwmd_setopt(pwm, PWMD_OPTION_INQUIRE_TOTAL,
+	    st.st_size ? st.st_size+strlen(inq.line) : 0);
+
+    if (rc)
+	return rc;
+
+    rc = pwmd_inquire(pwm, command, inquire_cb, &inq);
+    pwmd_free(inq.line);
+    return rc;
+}
+
+#ifdef HAVE_LIBREADLINE
+static int interactive_hook(void)
+{
+    process_cmd(pwm, NULL, 0, 1);
+    return 0;
+}
+
+static gpg_error_t do_interactive()
+{
+    gpg_error_t rc = 0;
+
+    fprintf(stderr, N_("WARNING: interactive mode doesn't use secure memory!\n"));
+    rl_initialize();
+    rl_event_hook = &interactive_hook;
+    rl_set_keyboard_input_timeout(10000);
+
+    for (;;) {
+	char *line;
+	char *result = NULL;
+
+	line = readline("pwm> ");
+
+	if (!line)
+	    break;
+	else if (!*line) {
+	    free(line);
+	    continue;
+	}
+
+#ifdef HAVE_READLINE_HISTORY
+	add_history(line);
+#endif
+
+	/* INQUIRE <COMMAND [options ...]> */
+	if (!strncasecmp(line, "inquire ", 8)) {
+	    fprintf(stderr, N_("Press CTRL-D to end. Press twice on a non-empty line.\n"));
+	    rc = send_inquire(STDIN_FILENO, line+8, NULL);
+	}
+	/* INQUIRE_FILE <FILENAME> <COMMAND [options ...]> */
+	else if (!strncasecmp(line, "inquire_file ", 13)) {
+	    char *p = strchr(line+13, ' ');
+	    char *f = p ? line+13 : NULL;
+	    int fd;
+
+	    if (!f || !*f || !*p) {
+		fprintf(stderr, N_("Syntax error. Usage: INQUIRE_FILE <filename> <command>\n"));
+		free(line);
+		continue;
+	    }
+
+	    f[strlen(f) - strlen(p++)] = 0;
+	    fd = open(f, O_RDONLY);
+
+	    if (fd == -1) {
+		fprintf(stderr, "%s\n", strerror(errno));
+		free(line);
+		continue;
+	    }
+	    else {
+		rc = send_inquire(fd, p, NULL);
+		close(fd);
+	    }
+	}
+	else
+	    rc = pwmd_command(pwm, &result, line);
+
+	free(line);
+
+	if (rc)
+	    fprintf(stderr, "ERR %i: %s\n", rc, pwmd_strerror(rc));
+	else if (result && *result)
+	    printf("%s%s", result, result[strlen(result)-1] != '\n' ? "\n" : "");
+
+	if (result)
+	    pwmd_free(result);
+    }
+
+    return 0;
+}
+#endif
+
 int main(int argc, char *argv[])
 {
     int opt;
+#ifdef HAVE_LIBREADLINE
+    int interactive = 0;
+#endif
     char *password = NULL;
+    char *keyfile = NULL;
+    int base64 = 0;
     char *filename = NULL;
     char *socketpath = NULL;
     char command[ASSUAN_LINELENGTH], *p;
@@ -287,11 +515,13 @@ int main(int argc, char *argv[])
     int show_status = 1;
     char *clientname = "pwmc";
     char *inquire = NULL;
+    char *inquire_line = NULL;
     long iter = -2;
     int have_iter = 0;
     int timeout = 0;
     int local_pin = 0;
     int force_save = 0;
+    char *cipher = NULL;
 #ifdef WITH_TCP
     char *host = NULL;
     int port = DEFAULT_PORT;
@@ -300,11 +530,13 @@ int main(int argc, char *argv[])
     char *known_hosts = NULL;
     int get = 0;
     int prot = PWMD_IP_ANY;
+    int use_agent = 0;
 #endif
     int tries = 0;
     int local_tries = 3;
     int try = 0;
     pwmd_socket_t s;
+    int lock_on_open = 1;
 #ifdef DEBUG
     int method = 0;
     fd_set rfds;
@@ -317,13 +549,17 @@ int main(int argc, char *argv[])
 #endif
 #ifdef WITH_TCP
 	OPT_HOST, OPT_PORT, OPT_IDENTITY, OPT_KNOWN_HOSTS, OPT_USER,
-	OPT_GET_HOSTKEY, OPT_IPV4, OPT_IPV6,
+	OPT_GET_HOSTKEY, OPT_IPV4, OPT_IPV6, OPT_USE_AGENT,
 #endif
 	OPT_URL, OPT_LOCAL, OPT_FORCE_SAVE, OPT_TTYNAME, OPT_TTYTYPE,
 	OPT_DISPLAY, OPT_LC_CTYPE, OPT_LC_MESSAGES, OPT_TIMEOUT, OPT_TRIES,
-	OPT_PINENTRY, OPT_PASSPHRASE, OPT_SOCKET, OPT_SAVE, OPT_ITERATIONS,
-	OPT_OUTPUT_FD, OPT_INQUIRE_FD, OPT_NO_STATUS, OPT_NAME, OPT_VERSION,
-	OPT_HELP,
+	OPT_PINENTRY, OPT_PASSPHRASE, OPT_KEYFILE, OPT_BASE64, OPT_SOCKET,
+	OPT_NOLOCK, OPT_SAVE, OPT_ITERATIONS, OPT_OUTPUT_FD, OPT_INQUIRE,
+	OPT_INQUIRE_FD, OPT_INQUIRE_LINE, OPT_NO_STATUS, OPT_NAME,
+	OPT_VERSION, OPT_HELP, OPT_CIPHER,
+#ifdef HAVE_LIBREADLINE
+	OPT_INTERACTIVE,
+#endif
     };
     const struct option long_opts[] = {
 #ifdef DEBUG
@@ -338,6 +574,7 @@ int main(int argc, char *argv[])
 	{ "get-hostkey", 0, 0, 'g' },
 	{ "ipv4", 0, 0, '4' },
 	{ "ipv6", 0, 0, '6' },
+	{ "use-agent", 0, 0, 0 },
 #endif
 	{ "url", 1, 0, 0 },
 	{ "local-pinentry", 0, 0 },
@@ -351,21 +588,30 @@ int main(int argc, char *argv[])
 	{ "tries", 1, 0, 0 },
 	{ "pinentry", 1, 0, 0 },
 	{ "passphrase", 1, 0, 'P' },
+	{ "key-file", 1, 0, 0 },
+	{ "base64", 0, 0, 0 },
 	{ "socket", 1, 0, 0 },
+	{ "no-lock", 0, 0, 0 },
 	{ "save", 0, 0, 'S' },
 	{ "iterations", 1, 0, 'I' },
 	{ "output-fd", 1, 0, 0 },
+	{ "inquire", 1, 0, 0 },
 	{ "inquire-fd", 1, 0, 0 },
+	{ "inquire-line", 1, 0, 'L' },
 	{ "no-status", 0, 0, 0 },
 	{ "name", 1, 0, 'n' },
 	{ "version", 0, 0, 0 },
 	{ "help", 0, 0, 0 },
+	{ "cipher", 1, 0, 0 },
+#ifdef HAVE_LIBREADLINE
+	{ "interactive", 0, 0 },
+#endif
 	{ 0, 0, 0, 0}
     };
 #ifdef WITH_TCP
-    const char *optstring = "46h:p:i:k:u:gy:t:d:P:I:Sn:";
+    const char *optstring = "L:46h:p:i:k:u:gy:t:d:P:I:Sn:";
 #else
-    const char *optstring = "y:t:d:P:I:Sn:";
+    const char *optstring = "L:y:t:d:P:I:Sn:";
 #endif
     int opt_index = 0;
 
@@ -374,6 +620,9 @@ int main(int argc, char *argv[])
     bindtextdomain("libpwmd", LOCALEDIR);
 #endif
 
+    if (!strcmp(basename(argv[0]), "pwmsh"))
+	interactive = 1;
+
     while ((opt = getopt_long(argc, argv, optstring, long_opts, &opt_index)) != -1) {
 	switch (opt) {
 	    /* Handle long options without a short option part. */
@@ -387,6 +636,20 @@ int main(int argc, char *argv[])
 			    method = 3;
 			break;
 #endif
+#ifdef WITH_TCP
+		    case OPT_USE_AGENT:
+			use_agent = 1;
+			break;
+#endif
+		    case OPT_KEYFILE:
+			keyfile = optarg;
+			break;
+		    case OPT_BASE64:
+			base64 = 1;
+			break;
+		    case OPT_NOLOCK:
+			lock_on_open = 0;
+			break;
 		    case OPT_URL:
 			url_string = optarg;
 			break;
@@ -412,6 +675,9 @@ int main(int argc, char *argv[])
 		    case OPT_SOCKET:
 			socketpath = pwmd_strdup(optarg);
 			break;
+		    case OPT_INQUIRE:
+			inquire = optarg;
+			break;
 		    case OPT_INQUIRE_FD:
 			inquirefd = atoi(optarg);
 			inquirefp = fdopen(inquirefd, "r");
@@ -448,14 +714,19 @@ int main(int argc, char *argv[])
 				"-PINENTRY "
 #endif
 #ifdef WITH_QUALITY
-				"+QUALITY "
+				"+CRACK "
 #else
-				"-QUALITY "
+				"-CRACK "
 #endif
 #ifdef MEM_DEBUG
 				"+MEM_DEBUG "
 #else
 				"-MEM_DEBUG "
+#endif
+#ifdef HAVE_LIBREADLINE
+				"+INTERACTIVE "
+#else
+				"-INTERACTIVE "
 #endif
 				"\n"
 				, PACKAGE_STRING, PACKAGE_BUGREPORT);
@@ -465,6 +736,14 @@ int main(int argc, char *argv[])
 			break;
 		    case OPT_HELP:
 			usage(argv[0], EXIT_SUCCESS);
+		    case OPT_CIPHER:
+			cipher = optarg;
+			break;
+#ifdef HAVE_LIBREADLINE
+		    case OPT_INTERACTIVE:
+			interactive = 1;
+			break;
+#endif
 		    default:
 			usage(argv[0], EXIT_FAILURE);
 		}
@@ -496,6 +775,9 @@ int main(int argc, char *argv[])
 		get = 1;
 		break;
 #endif
+	    case 'L':
+		inquire_line = optarg;
+		break;
 	    case 'y':
 		tty = optarg;
 		break;
@@ -525,11 +807,16 @@ int main(int argc, char *argv[])
 	}
     }
 
+    if (interactive && !isatty(STDIN_FILENO)) {
+	pwmd_free(password);
+	usage(argv[0], EXIT_FAILURE);
+    }
+
 #ifdef DEBUG
     if (!url_string) {
 #endif
 #ifdef WITH_TCP
-    if (host && !get && (!known_hosts || !ident)) {
+    if (host && !get && !ident) {
 	pwmd_free(password);
 	usage(argv[0], EXIT_FAILURE);
     }
@@ -551,7 +838,11 @@ int main(int argc, char *argv[])
 #endif
     
 #ifdef WITH_TCP
-    if (host) {
+    if (host || url_string) {
+#ifndef DEBUG
+	local_pin = 1;
+#endif
+
 	if (prot != PWMD_IP_ANY) {
 	    error = pwmd_setopt(pwm, PWMD_OPTION_IP_VERSION, prot);
 
@@ -559,6 +850,23 @@ int main(int argc, char *argv[])
 		goto done;
 	}
 
+	error = pwmd_setopt(pwm, PWMD_OPTION_KNOWNHOST_CB, knownhost_cb);
+
+	if (error)
+	    goto done;
+
+	error = pwmd_setopt(pwm, PWMD_OPTION_KNOWNHOST_DATA, clientname);
+
+	if (error)
+	    goto done;
+
+	error = pwmd_setopt(pwm, PWMD_OPTION_SSH_AGENT, use_agent);
+
+	if (error)
+	    goto done;
+
+	fprintf(stderr, N_("Connecting ...\n"));
+
 #ifdef DEBUG
 	if (method >= 2) {
 	    if (get) {
@@ -569,12 +877,12 @@ int main(int argc, char *argv[])
 		if (error)
 		    errx(EXIT_FAILURE, "%s: %s", host, pwmd_strerror(error));
 
-		error = process_cmd(pwm, &hostkey, 0);
+		error = process_cmd(pwm, &hostkey, 0, 0);
 
 		if (error)
 		    goto done;
 
-		printf("%s\n", hostkey);
+		printf("%s", hostkey);
 		pwmd_free(hostkey);
 		pwmd_free(password);
 		pwmd_close(pwm);
@@ -590,7 +898,7 @@ int main(int argc, char *argv[])
 	    if (error)
 		goto done;
 
-	    error = process_cmd(pwm, NULL, 0);
+	    error = process_cmd(pwm, NULL, 0, 0);
 
 	    if (error)
 		goto done;
@@ -605,7 +913,7 @@ int main(int argc, char *argv[])
 		if (error)
 		    goto done;
 
-		printf("%s\n", hostkey);
+		printf("%s", hostkey);
 		pwmd_free(hostkey);
 		pwmd_free(password);
 		pwmd_close(pwm);
@@ -636,15 +944,19 @@ int main(int argc, char *argv[])
     }
 #endif
 
+    fprintf(stderr, N_("Connected.\n"));
+    error = pwmd_setopt(pwm, PWMD_OPTION_PINENTRY_TITLE, NULL);
     error = pwmd_socket_type(pwm, &s);
 
     if (error)
 	goto done;
 
+#ifndef DEBUG
     if (s == PWMD_SOCKET_SSH && force_save && !local_pin) {
 	error = GPG_ERR_WRONG_KEY_USAGE;
 	goto done;
     }
+#endif
 
     if (have_iter) {
 	error = pwmd_command(pwm, &result, "VERSION");
@@ -661,6 +973,20 @@ int main(int argc, char *argv[])
 	}
     }
 
+    if (filename && lock_on_open) {
+	error = pwmd_setopt(pwm, PWMD_OPTION_LOCK_ON_OPEN, 1);
+
+	if (error)
+	    goto done;
+    }
+
+    if (base64) {
+	error = pwmd_setopt(pwm, PWMD_OPTION_BASE64, 1);
+
+	if (error)
+	    goto done;
+    }
+
     if (timeout > 0) {
 	error = pwmd_setopt(pwm, PWMD_OPTION_PINENTRY_TIMEOUT, timeout);
 
@@ -668,7 +994,7 @@ int main(int argc, char *argv[])
 	    goto done;
     }
 
-    if (local_pin && filename) {
+    if (local_pin && filename && !keyfile) {
 	error = pwmd_command(pwm, NULL, "ISCACHED %s", filename);
 
 	if (error == GPG_ERR_NOT_FOUND) {
@@ -769,6 +1095,37 @@ local_password:
 
 do_open:
     if (filename) {
+	fprintf(stderr, N_("Opening data file \"%s\" ...\n"), filename);
+
+#ifdef WITH_TCP
+	if (keyfile && (!local_pin || host || is_remote_url(url_string))) {
+	    struct inquire_s inq;
+
+	    memset(&inq, 0, sizeof(inq));
+	    inq.fd = open(keyfile, O_RDONLY);
+
+	    if (inq.fd == -1) {
+		error = gpg_error_from_syserror();
+		goto done;
+	    }
+
+	    inq.line = pwmd_calloc(1, ASSUAN_LINELENGTH);
+
+	    if (!inq.line) {
+		error = GPG_ERR_ENOMEM;
+		goto done;
+	    }
+
+	    error = pwmd_open_inquire(pwm, filename, inquire_cb, &inq);
+	    pwmd_free(inq.line);
+	    close(inq.fd);
+
+	    if (error)
+		goto done;
+
+	    goto read_command;
+	}
+#endif
 #ifdef DEBUG
 	switch (method) {
 	    case 0:
@@ -793,9 +1150,14 @@ do_open:
 	    goto done;
 
 	if (method >= 2)
-	    error = process_cmd(pwm, &result, 0);
+	    error = process_cmd(pwm, &result, 0, 0);
 #else
-	error = pwmd_open(pwm, filename);
+#ifdef WITH_TCP
+	if (host)
+	    error = pwmd_open2(pwm, filename);
+	else
+#endif
+	    error = pwmd_open(pwm, filename);
 
 	if (error && local_pin && error == GPG_ERR_INV_PASSPHRASE)
 	    goto local_password;
@@ -805,19 +1167,38 @@ do_open:
 	    goto done;
     }
 
-    if (filename) {
-	error = pwmd_command(pwm, &result, "LOCK");
+#ifdef WITH_TCP
+read_command:
+#endif
+#ifdef HAVE_LIBREADLINE
+    if (interactive) {
+	error = do_interactive();
 
 	if (error)
+	    goto do_exit;
+
+	goto done;
+    }
+#endif
+
+    if (inquire) {
+	int fd = fileno(inquirefp);
+
+	if (fd == -1) {
+	    error = gpg_error_from_syserror();
 	    goto done;
+	}
+
+	error = send_inquire(fd, inquire, inquire_line);
+	close(fd);
+	goto done;
     }
 
     fcntl(STDIN_FILENO, F_SETFL, O_NONBLOCK);
+    ssize_t n;
 
     for (;;) {
-	ssize_t n;
-
-	error = process_cmd(pwm, NULL, 1);
+	error = process_cmd(pwm, NULL, 1, 0);
 
 	if (error)
 	    goto done;
@@ -834,71 +1215,49 @@ do_open:
 	else if (!n)
 	    goto done;
 
-	command[n] = 0;
-
-	if (n && command[strlen(command)-1] == '\n')
-	    command[strlen(command)-1] = 0;
-
 	p = command;
+	command[n] = 0;
 	break;
     }
 
     if (!p || !*p)
 	goto done;
 
-    /*
-     * This is a known INQUIRE command. We use pwmd_inquire() to send the
-     * data from the do_inquire() callback function.
-     */
-    if (strncasecmp(p, "STORE ", 6) == 0) {
-	p += 6;
-	inquire = (char *)"STORE";
-    }
-    else if (strncasecmp(p, "IMPORT ", 7) == 0) {
-	p += 7;
-	inquire = (char *)"IMPORT";
-    }
-
-    if (inquire) {
-	struct inquire_s *inq = (struct inquire_s *)pwmd_malloc(sizeof(struct inquire_s));
-
-	if (!inq) {
-	    error = gpg_error_from_errno(ENOMEM);
-	    goto done;
-	}
-
-	inq->data = pwmd_strdup(p);
-	inq->fp = inquirefp;
-	error = pwmd_inquire(pwm, inquire, do_inquire, inq);
-	pwmd_free(inq);
-	goto done;
-    }
-
     if (strcasecmp(p, "BYE") == 0)
 	goto done;
 
     error = pwmd_command(pwm, &result, command);
-    memset(command, 0, sizeof(command));
 
     if (error)
 	goto done;
 
     if (result) {
 	fwrite(result, 1, strlen(result), outfp);
+	fflush(outfp);
 	pwmd_free(result);
     }
 
 done:
-    memset(command, 0, sizeof(command));
-    pwmd_free(password);
+    if (password)
+	pwmd_free(password);
+
     password = NULL;
 
     if (!error && save && filename) {
+	fprintf(stderr, N_("Saving changes ...\n"));
+
 	if (iter != -2) {
-	    error = pwmd_command(pwm, &result, "SET ITERATIONS=%i", iter);
+	    error = pwmd_setopt(pwm, PWMD_OPTION_ITERATIONS, iter);
 
 	    if (error)
-		goto done;
+		goto do_exit;
+	}
+
+	if (cipher) {
+	    error = pwmd_setopt(pwm, PWMD_OPTION_CIPHER, cipher);
+
+	    if (error)
+		goto do_exit;
 	}
 
 	if (local_pin) {
@@ -909,7 +1268,7 @@ again:
 
 		if (error && error != GPG_ERR_NOT_FOUND &&
 			error != GPG_ERR_ENOENT)
-		    goto done;
+		    goto do_exit;
 		else if (!error)
 		    goto do_save;
 	    }
@@ -917,14 +1276,14 @@ again:
 	    error = pwmd_getpin(pwm, filename, &p1, PWMD_PINENTRY_SAVE);
 
 	    if (error)
-		goto done;
+		goto do_exit;
 
 	    error = pwmd_getpin(pwm, filename, &password,
 		    PWMD_PINENTRY_SAVE_CONFIRM);
 
 	    if (error) {
 		pwmd_free(p1);
-		goto done;
+		goto do_exit;
 	    }
 
 	    if ((p1 || password) && ((!p1 && password) || (!password && p1) ||
@@ -944,50 +1303,79 @@ again:
 		pwmd_free(password);
 
 	    if (error)
-		goto done;
+		goto do_exit;
 	}
 
 	if (force_save) {
 	    error = pwmd_command(pwm, NULL, "CLEARCACHE %s", filename);
 
 	    if (error)
-		goto done;
+		goto do_exit;
 
 	    if (!local_pin) {
 		error = pwmd_setopt(pwm, PWMD_OPTION_PASSPHRASE, NULL);
 
 		if (error)
-		    goto done;
+		    goto do_exit;
 	    }
 	}
 
 do_save:
-#ifdef DEBUG
-	switch (method) {
-	    case 0:
-		error = pwmd_save(pwm);
-		break;
-	    case 1:
-		error = pwmd_save2(pwm);
-		break;
-	    case 2:
-		error = pwmd_save_async(pwm);
-		break;
-	    case 3:
-		error = pwmd_save_async2(pwm);
-		break;
-	}
+	if (keyfile && !local_pin) {
+	    struct inquire_s inq;
 
-	if (!error && method >= 2)
-	    error = process_cmd(pwm, NULL, 0);
+	    memset(&inq, 0, sizeof(inq));
+	    inq.fd = open(keyfile, O_RDONLY);
+
+	    if (inq.fd == -1) {
+		error = gpg_error_from_syserror();
+		goto do_exit;
+	    }
+
+	    inq.line = pwmd_calloc(1, ASSUAN_LINELENGTH);
+
+	    if (!inq.line) {
+		error = GPG_ERR_ENOMEM;
+		goto do_exit;
+	    }
+
+	    error = pwmd_save_inquire(pwm, inquire_cb, &inq);
+	    pwmd_free(inq.line);
+	    close(inq.fd);
+	}
+	else {
+#ifdef DEBUG
+	    switch (method) {
+		case 0:
+		    error = pwmd_save(pwm);
+		    break;
+		case 1:
+		    error = pwmd_save2(pwm);
+		    break;
+		case 2:
+		    error = pwmd_save_async(pwm);
+		    break;
+		case 3:
+		    error = pwmd_save_async2(pwm);
+		    break;
+	    }
+
+	    if (!error && method >= 2)
+		error = process_cmd(pwm, NULL, 0, 0);
 
 #else
-	error = pwmd_save(pwm);
+#ifdef WITH_TCP
+	    if (host)
+		error = pwmd_save2(pwm);
+	    else
 #endif
+		error = pwmd_save(pwm);
+#endif
+	}
     }
 
-    if (!error && filename)
-	error = pwmd_command(pwm, &result, "UNLOCK");
+do_exit:
+    memset(command, 0, sizeof(command));
 
     if (error) {
 	show_error(error);
